# Generated by Django 4.2.2 on 2023-07-05 19:12

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):
    dependencies = [
        ("diop", "0023_sitelog"),
    ]

    operations = [
        migrations.AlterField(
            model_name="sitelog",
            name="deliverygroup",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                to="diop.deliverygroup",
            ),
        ),
        migrations.AlterField(
            model_name="sitelog",
            name="machine",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                to="diop.machine",
            ),
        ),
        migrations.AlterField(
            model_name="sitelog",
            name="notification",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                to="diop.notification",
            ),
        ),
        migrations.AlterField(
            model_name="sitelog",
            name="reservation",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                to="diop.reservation",
            ),
        ),
        migrations.AlterField(
            model_name="sitelog",
            name="session",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                to="diop.session",
            ),
        ),
        migrations.AlterField(
            model_name="sitelog",
            name="task",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                to="diop.task",
            ),
        ),
        migrations.AlterField(
            model_name="sitelog",
            name="user",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                to="diop.user",
            ),
        ),
    ]



"""Diop Machine model."""

from loguru import logger as log

from django.db import models

from .deliverygroup import DeliveryGroup


class Machine(models.Model):
    """Class representing details about a Diop machine."""

    fqdn = models.CharField(
        "FQDN", max_length=64, primary_key=True, help_text="Fully Qualified Domain Name"
    )
    """**CharField [PRIMARY KEY]**

    The FQDN of the machine.
    """

    deliverygroup = models.ForeignKey(
        DeliveryGroup,
        on_delete=models.CASCADE,
        verbose_name="Delivery Group",
        help_text="Delivery Group the machine belongs to.",
    )
    """**Foreign Key**

    The `diop.models.deliverygroup.DeliveryGroup` associated to the machine.
    """

    state = models.CharField(
        "Machine State",
        max_length=16,
        help_text="Current machine status.",
    )
    """**CharField**

    The current status of the machine.
    """

    powerstate = models.CharField("Power State", max_length=16)
    """**CharField**

    The current power ⚡ state of the machine.
    """

    maintenance = models.BooleanField("Maintenance Mode", default=False)
    """**BooleanField**

    Flag indicating if the machine is currently in maintenance 🔧 mode.
    """

    registration = models.CharField("Registration State", max_length=16)
    """**CharField**

    The Citrix registration state of the machine.
    """

    agent = models.CharField("Agent Version", max_length=32, null=True)
    """**CharField**

    The version of the Citrix agent running on the machine.
    """

    active = models.BooleanField(
        "Active",
        default=True,
        help_text=(
            "Indicating if the machine is active or has been retired. Entries having "
            "this set to False will be ignored by update task completeness checks."
        ),
    )
    """**BooleanField**

    Flag indicating if the machine is active or has been retired. Entries having
    this set to False will be ignored by update task completeness checks.
    """

    updated = models.BooleanField(
        "Updated",
        default=True,
        help_text=(
            "Flag tracking if the machine status is up-to-date. Has to be set to False "
            "by every function updating an entry and only re-set to True once the "
            "update has completed successfully."
        ),
    )
    """**BooleanField**

    Flag for tracking if the machine status is up-to-date.

    Has to be set to `False`by every function updating an entry and only re-set
    to `True` once the update has completed successfully.
    """

    req_maintenance = models.BooleanField(
        "Request Maintenance",
        default=False,
        help_text="Enter maintenance mode as soon as the machine is free.",
    )
    """**BooleanField**

    Flag indicating maintenance mode should be activated as soon as possible.
    """

    def __str__(self):
        """Return the`hostname()` of the machine."""
        return self.hostname

    @property
    def hostname(self):
        """The machine's hostname, i.e. its `fqdn` until the first dot (`.`)."""
        return self.fqdn.split(".")[0]

    @classmethod
    def get_or_new(cls, fqdn: str, dg_name: str, state: str = ""):
        """Get a machine object from the DB or create a new one.

        Queries the DB for an existing machine object, looking for the tuple (fqdn,
        deliverygroup), corresponding to parameters `fqdn` and `dg_name`. If no such
        entry exists, a new one will be created.

        Parameters
        ----------
        fqdn : str
            The machine's FQDN, corresponds to the **primary key** of the
            `Machine` model.
        dg_name : str
            The machine's delivery group, used as a **foreign key** to reference
            the corresponding `diop.models.deliverygroup.DeliveryGroup` record.
        state : str, optional
            An optional state to set for the machine entry. Useful when querying for
            a machine object using session information (which also contains the
            machine's state).

        Returns
        -------
        diop.models.machine.Machine
        """
        fqdn = fqdn.lower()
        log.debug(f"🖥  Checking for machine [{fqdn}]...")

        if not state:
            state = "UNKNOWN"

        d_dg, _ = DeliveryGroup.get_or_new(dg_name)

        # query for the machine, passing in "state" to be used if a new record is
        # created (see below for *updating* the state on existing machines):
        d_machine, created = cls.objects.get_or_create(
            fqdn=fqdn,
            deliverygroup=d_dg,
            defaults={
                "state": state,
            },
        )

        if created:
            log.warning(f"🖥  Creating NEW machine: [{fqdn}] ✨")
        else:
            log.debug(f"🖥  Updating existing machine: [{fqdn}] 📝")

        # if the machine exists but "state" is differing we need to update (+ save)
        if d_machine.state != state:
            d_machine.state = state
            d_machine.save()

        return d_machine, created

    @classmethod
    def from_psy_session(cls, psy_session: dict):
        """Get a Machine object from a PSyTricks session dict.

        Parameters
        ----------
        psy_session : dict
            Session infos as returned by psytricks.ResTricksWrapper.get_sessions().

        Returns
        -------
        diop.models.machine.Machine
            The DB record corresponding to the session machine.
        """
        fqdn = psy_session["DNSName"].lower()
        dg_name = psy_session["DesktopGroupName"]
        state = psy_session["MachineSummaryState"]

        d_machine, _ = cls.get_or_new(fqdn, dg_name, state)

        return d_machine



# Generated by Django 4.2.1 on 2023-05-23 16:32

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("diop", "0013_alter_issue_description_alter_issue_resolved_and_more"),
    ]

    operations = [
        migrations.AddField(
            model_name="session",
            name="updated",
            field=models.BooleanField(
                default=True,
                help_text="Flag tracking if the session status is up-to-date. Has to be set to False by every function updating an entry and only re-set to True once the update has completed successfully.",
                verbose_name="Updated",
            ),
        ),
    ]



# Generated by Django 4.1.7 on 2023-04-17 12:38

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("diop", "0006_alter_deliverygroup_booking_required_and_more"),
    ]

    operations = [
        migrations.AddField(
            model_name="deliverygroup",
            name="access_users",
            field=models.ManyToManyField(to="diop.user"),
        ),
    ]



# Generated by Django 4.1.7 on 2023-04-14 10:04

from django.db import migrations, models


class Migration(migrations.Migration):
    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="SiteConfig",
            fields=[
                (
                    "setting",
                    models.CharField(max_length=64, primary_key=True, serialize=False),
                ),
                ("value", models.CharField(blank=True, max_length=512)),
            ],
        ),
        migrations.CreateModel(
            name="User",
            fields=[
                (
                    "username",
                    models.CharField(max_length=32, primary_key=True, serialize=False),
                ),
                ("fullname", models.CharField(max_length=128)),
                ("groupname", models.CharField(max_length=128)),
                ("department", models.CharField(max_length=128)),
                ("email", models.CharField(max_length=128)),
                ("ppms_group", models.CharField(max_length=64)),
                ("ppms_fullname", models.CharField(max_length=128)),
                ("gracetime", models.IntegerField(default=0)),
                ("active", models.BooleanField(default=True)),
                ("accessgroup_member", models.BooleanField(default=False)),
            ],
        ),
    ]



# Generated by Django 4.1.7 on 2023-04-17 15:37

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):
    dependencies = [
        ("diop", "0007_deliverygroup_access_users"),
    ]

    operations = [
        migrations.CreateModel(
            name="Session",
            fields=[
                (
                    "uid",
                    models.BigIntegerField(
                        help_text="Citrix:Uid (session identifier)",
                        primary_key=True,
                        serialize=False,
                        verbose_name="Session UID",
                    ),
                ),
                (
                    "t_start",
                    models.DateTimeField(
                        help_text="Citrix:StartTime (session start timestamp)",
                        verbose_name="Start Time",
                    ),
                ),
                (
                    "t_end",
                    models.DateTimeField(
                        blank=True,
                        help_text="Timestamp when the end of the session was detected first.",
                        null=True,
                        verbose_name="End Time",
                    ),
                ),
                (
                    "fqdn",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="diop.machine",
                        verbose_name="Host Name (FQDN)",
                    ),
                ),
                (
                    "user",
                    models.ForeignKey(
                        help_text="User associated to the session, can be empty!",
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        to="diop.user",
                        verbose_name="User",
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="SessionProperties",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "t_change",
                    models.DateTimeField(
                        help_text="Citrix:SessionStateChangeTime",
                        verbose_name="State Change Time",
                    ),
                ),
                (
                    "state",
                    models.CharField(max_length=16, verbose_name="Session State"),
                ),
                (
                    "client_address",
                    models.CharField(
                        help_text="IP address of the client currently (or last) connected.",
                        max_length=46,
                        verbose_name="Client IP",
                    ),
                ),
                (
                    "client_name",
                    models.CharField(
                        help_text="The name of the client currently (or last) connected.",
                        max_length=64,
                        verbose_name="Client Name",
                    ),
                ),
                (
                    "client_version",
                    models.CharField(
                        help_text="The version of the client currently (or last) connected.",
                        max_length=16,
                        verbose_name="Client Version",
                    ),
                ),
                (
                    "session",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="diop.session",
                        verbose_name="Session UID",
                    ),
                ),
            ],
            options={
                "verbose_name": "Session Properties",
                "verbose_name_plural": "Session Properties",
            },
        ),
    ]



# Generated by Django 4.2.2 on 2023-06-13 15:55

from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone


class Migration(migrations.Migration):
    dependencies = [
        ("diop", "0015_rename_session_fqdn_session_machine"),
    ]

    operations = [
        migrations.CreateModel(
            name="Notification",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "method",
                    models.CharField(
                        choices=[("P", "Pop Up"), ("E", "Email")],
                        default="P",
                        help_text="How the notification should be delievered to the user.",
                        max_length=1,
                        verbose_name="Delivery Method",
                    ),
                ),
                (
                    "cc_adm",
                    models.BooleanField(
                        help_text="Flag if admins should be notified as well.",
                        verbose_name="Cc Operators",
                    ),
                ),
                (
                    "subject",
                    models.CharField(
                        max_length=256, verbose_name="The notification subject / title."
                    ),
                ),
                (
                    "body",
                    models.CharField(
                        max_length=8192, verbose_name="The notification body / content."
                    ),
                ),
                (
                    "t_created",
                    models.DateTimeField(
                        default=django.utils.timezone.now,
                        help_text="Notification creation timestamp.",
                        verbose_name="Creation Time",
                    ),
                ),
                (
                    "t_sent",
                    models.DateTimeField(
                        help_text="Notification delivery timestamp (empty if not yet delieverd).",
                        null=True,
                        verbose_name="Sent Time",
                    ),
                ),
                (
                    "creator",
                    models.CharField(
                        blank=True,
                        help_text="Creator / origin of the notification.",
                        max_length=64,
                        verbose_name="Creator",
                    ),
                ),
                (
                    "session",
                    models.ForeignKey(
                        blank=True,
                        help_text="Session the notification relates to.",
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        to="diop.session",
                    ),
                ),
                (
                    "user",
                    models.ForeignKey(
                        help_text="User which the notification is addressed to.",
                        on_delete=django.db.models.deletion.CASCADE,
                        to="diop.user",
                    ),
                ),
            ],
        ),
    ]



"""Populate the Diop database with provided fixtures."""

from glob import glob
from pathlib import Path

from django.core.management.base import BaseCommand, CommandError
from django.core.management.commands import loaddata


class Command(BaseCommand):
    """Command to load provided fixtures into the Diop database.

    This is a convenience wrapper to facilitate the initial setup and simplify
    the installation commands. All it does is to figure out the path to the
    provided YAML fixtures and call the `handle()` method on the `loaddata`
    management command object with the appropriate parameters.
    """

    help = "Populate the Diop database with provided fixtures"

    def handle(self, *args, **options):
        """Entry point method for the management command."""
        fixt_dir = Path(__file__).parent.parent.parent / "fixtures"
        fixtures = glob(str(fixt_dir / "*.yaml"))

        # the minimal options dict for calling handle() below:
        lf_options = {
            "verbosity": 3,
            "database": "default",
            "app_label": None,
            "ignore": False,
            "exclude": [],
            "format": "yaml",
        }

        loadfixtures = loaddata.Command()
        loadfixtures.handle(*fixtures, **lf_options)



"""Django settings for the `diop_core` project.

Configuration is done using a file defined in the environment variable
`SETTINGS_ENV` (that variable has to be present when launching the Django
application). Reading the file and processing the settings is implemented
through the [django-environ][dj-environ] project.

Example
-------
To change a configuration value, the settings file needs to contain an
environment variable using the identical name as documented here, e.g. to
configure the Diop LDAP authentication bind password expected in the variable
diop_core.settings.AUTH_LDAP_BIND_PASSWORD, a line like this will be required in
the settings file:

`AUTH_LDAP_BIND_PASSWORD=SuP3rS3kr3T`

Notes
-----
The API docs generated by [pdoc][pdoc] will only list those items that are
configured through the settings file. Items that are not intended to be changed
(and thus are hard-coded in diop_core.settings) will be omitted.

References
----------
For generic information on this file, see:

* https://docs.djangoproject.com/en/4.1/topics/settings/
* https://docs.djangoproject.com/en/4.1/ref/settings/

[dj-environ]: https://django-environ.readthedocs.io/
[pdoc]: https://pdoc.dev
"""

# ruff: noqa: E501 (line length)
# ruff: noqa: F401 (imported but unused)

import os
from pathlib import Path

import ldap
from django_auth_ldap.config import LDAPSearch

import environ


env = environ.Env(
    # set casting (=default values in case they're not set in the environment)
    DEBUG=(bool, False)
)
"""@private (skip in pdoc)"""


# read the file defined by the environment variable `SETTINGS_ENV`:
env.read_env(env.str("SETTINGS_ENV"))


SECRET_KEY = env.str("SECRET_KEY")

SITE_ID = 1
"""@private (skip in pdoc)"""


DEBUG = env.bool("DEBUG")
"""[Django debug setting][djdoc-debug],
**DON'T RUN WITH DEBUG TURNED ON IN PRODUCTION**!

[djdoc-debug]: https://docs.djangoproject.com/en/4.1/ref/settings/#debug
"""


# allowed hosts may be overridden in the private settings!
# having a pseudo-hostname here is handy during development (can be pointed to
# wherever using a corresponding entry in /etc/hosts)
ALLOWED_HOSTS = env.list(
    "ALLOWED_HOSTS", default=["127.0.0.1", "localhost", "diop.test"]
)

# required e.g. if running behind an nginx reverse proxy:
SECURE_PROXY_SSL_HEADER = ("HTTP_X_FORWARDED_PROTO", "https")
"""@private (skip in pdoc)"""


# Application definition

INSTALLED_APPS = [
    "diop",
    # also use whitenoise in development (when started via `runserver`):
    "whitenoise.runserver_nostatic",
    "django.contrib.admin",
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "django.contrib.humanize",
    "django.contrib.sessions",
    "django.contrib.sites",
    "django.contrib.messages",
    "django.contrib.staticfiles",
    "django_q",
    "django_extensions",
]
"""@private (skip in pdoc)"""

MIDDLEWARE = [
    "django.middleware.security.SecurityMiddleware",
    "whitenoise.middleware.WhiteNoiseMiddleware",
    "django.contrib.sessions.middleware.SessionMiddleware",
    "django.middleware.common.CommonMiddleware",
    "django.middleware.csrf.CsrfViewMiddleware",
    "django.contrib.auth.middleware.AuthenticationMiddleware",
    "django.contrib.messages.middleware.MessageMiddleware",
    "django.middleware.clickjacking.XFrameOptionsMiddleware",
    "django.contrib.sites.middleware.CurrentSiteMiddleware",
]
"""@private (skip in pdoc)"""

ROOT_URLCONF = "diop_core.urls"
"""@private (skip in pdoc)"""

TEMPLATES = [
    {
        "BACKEND": "django.template.backends.django.DjangoTemplates",
        "DIRS": [],
        "APP_DIRS": True,
        "OPTIONS": {
            "context_processors": [
                "django.template.context_processors.debug",
                "django.template.context_processors.request",
                "django.contrib.auth.context_processors.auth",
                "django.contrib.messages.context_processors.messages",
            ],
        },
    },
]
"""@private (skip in pdoc)"""

WSGI_APPLICATION = "diop_core.wsgi.application"
"""@private (skip in pdoc)"""


# Database
# https://docs.djangoproject.com/en/4.1/ref/settings/#databases

DATABASES = {"default": env.db("DATABASE_URL")}
"""The [backend database][djdoc-db]. For Diop, only MySQL is tested.

[djdoc-db]: https://docs.djangoproject.com/en/4.1/ref/settings/#databases
"""

# Password validation
# https://docs.djangoproject.com/en/4.1/ref/settings/#auth-password-validators

AUTH_PASSWORD_VALIDATORS = [
    {
        "NAME": "django.contrib.auth.password_validation.UserAttributeSimilarityValidator",
    },
    {
        "NAME": "django.contrib.auth.password_validation.MinimumLengthValidator",
    },
    {
        "NAME": "django.contrib.auth.password_validation.CommonPasswordValidator",
    },
    {
        "NAME": "django.contrib.auth.password_validation.NumericPasswordValidator",
    },
]
"""@private (skip in pdoc)"""

LOGGING = {
    "version": 1,
    "disable_existing_loggers": False,
    "handlers": {"console": {"class": "logging.StreamHandler"}},
    "loggers": {"django_auth_ldap": {"level": "DEBUG", "handlers": ["console"]}},
}
"""@private (skip in pdoc)"""

AUTHENTICATION_BACKENDS = [
    "django.contrib.auth.backends.ModelBackend",
    "django_auth_ldap.backend.LDAPBackend",
]
"""@private (skip in pdoc)"""

AUTH_LDAP_BIND_DN = env.str("AUTH_LDAP_BIND_DN")
AUTH_LDAP_BIND_PASSWORD = env.str("AUTH_LDAP_BIND_PASSWORD")
AUTH_LDAP_SERVER_URI = env.str("AUTH_LDAP_SERVER_URI")
AUTH_LDAP_BASE_DN = env.str("AUTH_LDAP_BASE_DN")


AUTH_LDAP_GLOBAL_OPTIONS = {
    ldap.OPT_REFERRALS: 0,
    ldap.OPT_PROTOCOL_VERSION: ldap.VERSION3,
    ldap.OPT_X_TLS_REQUIRE_CERT: ldap.OPT_X_TLS_ALLOW,
    ## OPT_X_TLS_NEWCTX doesn't seem to be required:
    # ldap.OPT_X_TLS_NEWCTX: 0,
    ## set OPT_DEBUG_LEVEL to a value bigger than 0 to get debug messages:
    # ldap.OPT_DEBUG_LEVEL: 0,
}
"""@private (skip in pdoc)"""

## NOTE: using the `AUTH_LDAP_CONNECTION_OPTIONS` didn't seem to do the job,
## hence we moved to the `AUTH_LDAP_GLOBAL_OPTIONS` above!
# AUTH_LDAP_CONNECTION_OPTIONS: {
#     ldap.OPT_REFERRALS: 0,
#     # ldap.OPT_PROTOCOL_VERSION: ldap.VERSION3,
#     # ldap.OPT_X_TLS_REQUIRE_CERT: ldap.OPT_X_TLS_ALLOW,
#     # ldap.OPT_X_TLS_NEWCTX: 0,
# }
"""@private (skip in pdoc)"""

AUTH_LDAP_USER_SEARCH = LDAPSearch(
    AUTH_LDAP_BASE_DN,
    ldap.SCOPE_SUBTREE,
    "sAMAccountName=%(user)s",
)
"""@private (skip in pdoc)"""

AUTH_LDAP_USER_ATTR_MAP = {
    "username": "sAMAccountName",
    "first_name": "givenName",
    "last_name": "sn",
    "email": "mail",
}
"""@private (skip in pdoc)"""

# Internationalization
# https://docs.djangoproject.com/en/4.1/topics/i18n/

LANGUAGE_CODE = "en-us"
"""@private (skip in pdoc)"""

DATE_FORMAT = "Y-m-d"
"""@private (skip in pdoc)"""

DATETIME_FORMAT = "Y-m-d H:i:s"
"""@private (skip in pdoc)"""

# Disable localized "L10N" formatting of data, otherwise the date and time
# formats defined above will be ignored / overridden.
# NOTE: Django 5.x will always have this set to True, currently it is unclear
# how this will play together with explicitly set date and time formats!
USE_L10N = False
"""@private (skip in pdoc)"""

TIME_ZONE = "Europe/Zurich"
"""@private (skip in pdoc)"""

USE_I18N = True
"""@private (skip in pdoc)"""

USE_TZ = True
"""@private (skip in pdoc)"""


# Static files (CSS, JavaScript, Images)
# https://docs.djangoproject.com/en/4.1/howto/static-files/


BASE_DIR = Path(__file__).resolve().parent.parent
"""The base directory of the project, i.e. where the `diop` and `diop_core`
directories can be found.

To build paths inside the project, use it like this: `BASE_DIR / 'subdir'`.
"""
"""@private (skip in pdoc)"""


### REVIEW! ###
STATIC_URL = "static/"
STATIC_ROOT = BASE_DIR / "staticfiles"
"""The absolute path where `collectstatic` will collect static files for deployment.

https://docs.djangoproject.com/en/4.1/ref/settings/#static-root
"""
"""@private (skip in pdoc)"""


STATICFILES_DIRS = [BASE_DIR / "diop" / "static"]
"""Additional locations the staticfiles app will traverse, e.g. using the
`collectstatic` or `findstatic` management command or through the static file
serving view. Should be set to a list of strings that contain full paths to
the additional files and directories.

https://docs.djangoproject.com/en/4.1/ref/settings/#staticfiles-dirs
"""
"""@private (skip in pdoc)"""


# STATICFILES_STORAGE = "whitenoise.storage.CompressedManifestStaticFilesStorage"
STORAGES = {
    "staticfiles": {
        "BACKEND": "whitenoise.storage.CompressedManifestStaticFilesStorage",
    },
}
"""@private (skip in pdoc)"""


# Default primary key field type
# https://docs.djangoproject.com/en/4.1/ref/settings/#default-auto-field

DEFAULT_AUTO_FIELD = "django.db.models.BigAutoField"
"""@private (skip in pdoc)"""

CACHES = {
    "default": {
        "BACKEND": "django.core.cache.backends.db.DatabaseCache",
        "LOCATION": "django_diop_core_cache_table",
    }
}
"""@private (skip in pdoc)"""

# Email configuration
# https://docs.djangoproject.com/en/4.2/topics/email/
EMAIL_BACKEND = env.str("EMAIL_BACKEND")
EMAIL_HOST = env.str("EMAIL_HOST")


# django-q2 ORM broker connection
Q_CLUSTER = {
    "name": "DjangORM",
    "label": "Task Queue",
    "workers": 1,
    "timeout": 60,
    "retry": 120,
    "catch_up": False,
    "save_limit": 1000,
    "save_limit_per": "func",
    # "queue_limit": 50,
    # "bulk": 10,
    "orm": "default",
    "ALT_CLUSTERS": {
        "statusupdates": {
            "timeout": 30,
            "retry": 99999,
            "max_attempts": 1,
        },
        "housekeeping": {
            "timeout": 900,
            "retry": 99999,
            "workers": 2,
            "max_attempts": 1,
        },
        "activedirectory": {
            "timeout": 300,
            "retry": 600,
            "max_attempts": 4,
        },
        "ppms-short": {
            "timeout": 30,
            "retry": 120,
            "max_attempts": 4,
        },
        "ppms-long": {
            "timeout": 900,
            "retry": 1200,
            "max_attempts": 4,
        },
    },
}
"""@private (skip in pdoc)"""


PYPPMS_API_KEY = env.str("PYPPMS_API_KEY")
PYPPMS_URI = env.str("PYPPMS_URI")
PYPPMS_CACHE = env.str("PYPPMS_CACHE")
PYPPMS_CORE_REF = env.str("PYPPMS_CORE_REF")
PYPPMS_LOC_FILTER = env.str("PYPPMS_LOC_FILTER")
PYPPMS_NAME_GROUPS = env.list("PYPPMS_NAME_GROUPS")


PSYTRICKS_BASE_URL = env.str("PSYTRICKS_BASE_URL", default="http://localhost:8080/")
"""The address where to connect to the ResTricks service.

Used as the `base_url` parameter for the `psytricks.wrapper.ResTricksWrapper`
object. In case not specified, connecting via `localhost` will be attempted.
"""
if not PSYTRICKS_BASE_URL[-1] == "/":
    PSYTRICKS_BASE_URL += "/"

PSYTRICKS_VERIFY = env.bool("PSYTRICKS_VERIFY", default=True)
"""Request validating the ResTricks server version upon connecting.

By setting this to `False`, a mismatch of the server-side version may be
ignored.
"""


DIOP_USER_AD_GROUP = env.str("DIOP_USER_AD_GROUP")
DIOP_SUPPORT_CONTACT = env.str("DIOP_SUPPORT_CONTACT")
DIOP_DISCONNECTED_MAX = env.int("DIOP_DISCONNECTED_MAX")
DIOP_NTDOMAIN = env.str("DIOP_NTDOMAIN")


DIOP_GRACEFUL = env.bool("DIOP_GRACEFUL")
"""When running in "graceful" mode, some tasks operate in a less impacting way.

For example, removing access to a delivery group will be done such that the
standard group will still be allowed to connect.

Most likely not useful in production!
"""


DIOP_DRY_RUN = env.bool("DIOP_DRY_RUN")
"""Running in "dry-run" mode will prevent state-changes to the CVAD platform.

For example changing access permissions, terminating sessions or sending pop-up
messages will NOT be effective, instead only a log message will be issued.

It will NOT affect components fetching status information to update the DB -
they will run normally (as they do not change the CVAD platform). Actions that
are skipped due to dry-run being active will print a warning-level message
to the log instead.
"""



# Generated by Django 4.2.2 on 2023-06-13 21:50

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("diop", "0017_alter_machine_deliverygroup_alter_machine_state"),
    ]

    operations = [
        migrations.AddField(
            model_name="machine",
            name="req_maintenance",
            field=models.BooleanField(
                default=False,
                help_text="Enter maintenance mode as soon as the machine is free.",
                verbose_name="Request Maintenance",
            ),
        ),
    ]



"""Diop models."""

# ruff: noqa: F401 (imported but unused)

from .user import User
from .deliverygroup import DeliveryGroup
from .machine import Machine
from .session import Session
from .session_properties import SessionProperties
from .task import Task
from .issue import Issue
from .notification import Notification
from .reservation import Reservation
from .ppms_booking import PpmsBooking
from .site_log import SiteLog
from .message_template import MessageTemplate



# Generated by Django 4.1.7 on 2023-04-14 11:40

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("diop", "0002_machine"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="siteconfig",
            options={"verbose_name": "Site Config"},
        ),
        migrations.AlterField(
            model_name="machine",
            name="fqdn",
            field=models.CharField(
                max_length=64,
                primary_key=True,
                serialize=False,
                verbose_name="Fully Qualified Domain Name",
            ),
        ),
        migrations.AlterField(
            model_name="machine",
            name="maintenance",
            field=models.BooleanField(default=False, verbose_name="Maintenance Mode"),
        ),
        migrations.AlterField(
            model_name="machine",
            name="powerstate",
            field=models.CharField(max_length=16, verbose_name="Power State"),
        ),
        migrations.AlterField(
            model_name="machine",
            name="registration",
            field=models.CharField(max_length=16, verbose_name="Registration State"),
        ),
        migrations.AlterField(
            model_name="machine",
            name="state",
            field=models.CharField(max_length=16, verbose_name="Machine State"),
        ),
    ]



# Generated by Django 4.2.2 on 2023-06-21 12:31

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):
    dependencies = [
        ("diop", "0020_reservation_ppmsbooking_and_more"),
    ]

    operations = [
        migrations.AlterField(
            model_name="issue",
            name="description",
            field=models.TextField(),
        ),
        migrations.AlterField(
            model_name="issue",
            name="task",
            field=models.ForeignKey(
                blank=True,
                help_text="Related task causing the issue (if applicable).",
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                to="diop.task",
            ),
        ),
        migrations.AlterField(
            model_name="issue",
            name="user",
            field=models.ForeignKey(
                blank=True,
                help_text="Related user (if applicable).",
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                to="diop.user",
            ),
        ),
    ]



"""Diop Task model."""

from django.db import models
from django.utils.timezone import now


class Task(models.Model):
    name = models.CharField("Task Name", max_length=128, primary_key=True)
    # NOTE: it is on purpose that `t_start` is not using `auto_now_add` since this
    # allows to set the start timestamp explicitly - this is needed when a task is
    # started again at a later point (which is standard behavior):
    t_start = models.DateTimeField("Start Time", default=now)
    t_end = models.DateTimeField(
        "End Time",
        null=True,
        blank=True,
        help_text="Empty if task is still running or did not succeed.",
    )
    unfinished = models.SmallIntegerField(
        default=0,
        help_text="Counter on previous unfinished runs when next one tries to start.",
    )
    failed = models.SmallIntegerField(
        default=0, help_text="Counter on previously failed attempts."
    )
    description = models.TextField(null=False, blank=True)

    def __str__(self):
        return self.name

    @property
    def short(self):
        """Return the 'short' task name, i.e. everything following the last dot.

        Returns
        -------
        str
        """
        return self.name.split(".")[-1]



# Generated by Django 4.2.1 on 2023-06-02 18:48

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):
    dependencies = [
        ("diop", "0014_session_updated"),
    ]

    operations = [
        migrations.RenameField(
            model_name="session",
            old_name="fqdn",
            new_name="machine",
        ),
        migrations.AlterField(
            model_name="session",
            name="machine",
            field=models.ForeignKey(
                help_text="Machine associated to the session",
                on_delete=django.db.models.deletion.CASCADE,
                to="diop.machine",
                verbose_name="Machine",
            ),
        ),
    ]



# Generated by Django 4.2.2 on 2023-07-03 13:26

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):
    dependencies = [
        ("diop", "0022_alter_notification_body_alter_task_description"),
    ]

    operations = [
        migrations.CreateModel(
            name="SiteLog",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.CharField(max_length=128, verbose_name="Name")),
                (
                    "parameters",
                    models.CharField(max_length=512, verbose_name="Parameters"),
                ),
                (
                    "time",
                    models.DateTimeField(
                        auto_now_add=True, help_text="Time the action was triggered."
                    ),
                ),
                ("details", models.TextField()),
                (
                    "deliverygroup",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        to="diop.deliverygroup",
                    ),
                ),
                (
                    "machine",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        to="diop.machine",
                    ),
                ),
                (
                    "notification",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        to="diop.notification",
                    ),
                ),
                (
                    "reservation",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        to="diop.reservation",
                    ),
                ),
                (
                    "session",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        to="diop.session",
                    ),
                ),
                (
                    "task",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        to="diop.task",
                    ),
                ),
                (
                    "user",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        to="diop.user",
                    ),
                ),
            ],
            options={
                "verbose_name": "Site Log",
            },
        ),
    ]



# Generated by Django 4.2.2 on 2023-07-07 08:31

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("diop", "0025_alter_user_department_alter_user_fullname_and_more"),
    ]

    operations = [
        migrations.RemoveField(
            model_name="user",
            name="active",
        ),
        migrations.AddField(
            model_name="user",
            name="enabled",
            field=models.BooleanField(
                default=True,
                help_text="Account is enabled in AD (2nd bit of `userAccountControl` is 0).",
            ),
        ),
    ]



"""Create Diop systemd service unit files."""

from pathlib import Path
from textwrap import dedent
import sys

from django.core.management.base import BaseCommand, CommandError

from diop_core.settings import env


_SETTINGS_ENV = Path(env.str("SETTINGS_ENV")).resolve()
_VENV_BIN = Path(sys.executable).parent
# NOTE: Path(env.str("VIRTUAL_ENV")) fails when venv not activated


class Command(BaseCommand):
    """Command to create systemd service unit files."""

    help = "Create Diop systemd service unit files"

    @staticmethod
    def q2_unit(desc, cluster, workingdir=None, loglevel="SUCCESS"):
        """Render a systemd unit file content for a Django-Q2 worker.

        Parameters
        ----------
        desc : str
            The description to put into the `Unit.Description` field. Will be
            prefixed with `Diop Task Queue - `.
        cluster : str
            The name of the cluster that the worker started by this unit file
            will belong to (used for the `Q_CLUSTER_NAME` environment variable).
        workingdir : str, optional
            The value to set for `Service.WorkingDirectory`, by default
            `/home/django`.
        loglevel : str, optional
            The loguru loglevel name, by default `SUCCESS`.

        Returns
        -------
        str
        """
        workingdir = workingdir if workingdir else "/home/django"

        content = f"""
        [Unit]
        Description=Diop Task Queue - {desc}
        After=network.target

        [Service]
        User=django
        WorkingDirectory={workingdir}
        Environment=SETTINGS_ENV={_SETTINGS_ENV}
        Environment=LOGURU_LEVEL={loglevel}
        Environment=Q_CLUSTER_NAME={cluster}
        ExecStartPre={_VENV_BIN}/diop-manage migrate
        ExecStart={_VENV_BIN}/diop-manage qcluster

        [Install]
        WantedBy=multi-user.target
        """
        return dedent(content[1:])

    @staticmethod
    def diop_web_unit(workers=2, bind=None, workingdir=None, loglevel="SUCCESS"):
        """Render the systemd unit file content for the Diop gunicorn service.

        Parameters
        ----------
        workers : int, optional
            Number of gunicorn workers, by default 2
        bind : str, optional
            The bind address and port for gunicorn, will use `0.0.0.0:8000` in
            case it's not specified.
        workingdir : str, optional
            The value to set for `Service.WorkingDirectory`, by default
            `/home/django`.
        loglevel : str, optional
            The loguru loglevel name, by default `SUCCESS`.

        Returns
        -------
        str
        """
        workingdir = workingdir if workingdir else "/home/django"
        bind = bind if bind else "0.0.0.0:8000"

        content = f"""
        [Unit]
        Description=Diop Gunicorn web service
        After=network.target

        [Service]
        User=django
        WorkingDirectory={workingdir}
        Environment=SETTINGS_ENV={_SETTINGS_ENV}
        Environment=LOGURU_LEVEL={loglevel}
        ExecStartPre={_VENV_BIN}/diop-manage migrate
        ExecStart={_VENV_BIN}/gunicorn --workers {workers} --bind {bind} diop_core.wsgi

        [Install]
        WantedBy=multi-user.target
        """
        return dedent(content[1:])

    @staticmethod
    def write_units(outdir):
        """Write all unit files into the given directory.

        The target directory will be created in case it's not present, existing
        unit files will NOT be overwritten!

        Parameters
        ----------
        outdir : str or Path
            Where to create the unit files.
        """

        def write_file(fpath: Path, content: str):
            if fpath.exists():
                print(f"ERROR: file [{fpath}] already exists, skipping!")
                return
            fpath.write_text(content)
            print(fpath)

        outdir = Path(outdir).resolve()
        outdir.mkdir(exist_ok=True)

        print("Writing systemd unit files:")
        unitfile = outdir / "diop-web.service"
        write_file(unitfile, Command.diop_web_unit())

        q2_units = [
            ("activedirectory", "ActiveDirectory Tasks (5m)"),
            ("housekeeping", "Housekeeping Tasks (15m)"),
            ("ppms-long", "PPMS tasks LONG (15m)"),
            ("ppms-short", "PPMS tasks SHORT (30s)"),
            ("statusupdates", "Status Update Tasks"),
        ]
        for unit in q2_units:
            cluster = unit[0]
            desc = unit[1]
            unitfile = outdir / f"diop-q2-{cluster}.service"
            write_file(unitfile, Command.q2_unit(desc, cluster))

    def add_arguments(self, parser):
        """Add commandline arguments to the parser."""
        parser.add_argument(
            "--outdir",
            type=str,
            required=True,
            help="Directory where to store the unit files.",
        )

    def handle(self, *args, **options):
        """Entry point method for the management command."""
        Command.write_units(options["outdir"])



"""Diop Session model."""

from datetime import timedelta

from loguru import logger as log

from django.db import models
from django.utils.timezone import make_aware, now

from .machine import Machine
from .user import User


class Session(models.Model):
    """Class representing details about a Diop session."""

    uid = models.BigIntegerField(
        "Session UID",
        primary_key=True,
        help_text="Citrix:Uid (session identifier)",
    )
    """**BigIntegerField [PRIMARY KEY]**

    The *Session UID* as given by Citrix / CVAD.
    """

    machine = models.ForeignKey(
        Machine,
        on_delete=models.CASCADE,
        verbose_name="Machine",
        help_text="Machine associated to the session",
    )
    """**Foreign Key**

    The `diop.models.machine.Machine` associated to the session.
    """

    user = models.ForeignKey(
        User,
        on_delete=models.CASCADE,
        verbose_name="User",
        null=True,
        blank=True,
        help_text="User associated to the session, can be empty!",
    )
    """**Foreign Key**

    The `diop.models.user.User` associated to the session, **may be empty!**
    """

    t_start = models.DateTimeField(
        "Start Time", help_text="Citrix:StartTime (session start timestamp)"
    )
    """**DateTimeField**

    The session start time as provided by Citrix / CVAD.
    """

    t_end = models.DateTimeField(
        "End Time",
        null=True,
        blank=True,
        help_text="Timestamp when the end of the session was detected first.",
    )
    """**DateTimeField**

    The timestamp when the end of the session was detected by Diop.

    **Important:** this information is **derived** by Diop, as it is **not
    provided by Citrix / CVAD** (at least not reliably)!
    """

    updated = models.BooleanField(
        "Updated",
        default=True,
        help_text=(
            "Flag tracking if the session status is up-to-date. Has to be set to False "
            "by every function updating an entry and only re-set to True once the "
            "update has completed successfully."
        ),
    )
    """**BooleanField**

    Flag tracking if the session status is up-to-date. Must be set to `False` by
    every function updating an entry and only re-set to `True` once the update
    has completed successfully.
    """

    def __str__(self):
        """Get a string representation of the session."""
        username = "-"
        if self.user:
            username = self.user.username
        return f"{str(self.uid)} [{username}]"

    @property
    def cur_properties(self):
        """Get the newest session properties record of this session.

        Returns
        -------
        diop.models.session_properties.SessionProperties
        """
        return self.sessionproperties_set.order_by("t_change").last()

    @property
    def cur_state(self):
        """Get the *state* of this session record.

        Returns
        -------
        str
            Either the string `ended` in case this session record's `t_end` is
            not empty, or otherwise the state of the last corresponding record
            in `diop.models.session_properties.SessionProperties.state`.
        """
        if self.t_end:
            return "ended"

        return self.sessionproperties_set.order_by("t_change").last().state

    @property
    def disconnected_since(self):
        """Number of seconds since the session is in "disconnected" state.

        Returns
        -------
        int
            The number of seconds since the session was disconnected, or -1 in
            case the session is active.
        """
        session_props = self.cur_properties
        if session_props.state != "disconnected":
            return -1

        delta = int((now() - session_props.t_change).total_seconds())
        log.trace(f"Session [{self}] is disconnected since {delta}s")

        return delta

    @property
    def prospected_termination(self):
        """Expected termination timepoint for disconnected sessions.

        Please note that this does NOT take into account the actual state of the
        session, it simply adds the user-specific "disconnected_max" timespan to
        the last session status change.

        Returns
        -------
        datetime.datetime
        """
        return self.cur_properties.t_change + timedelta(
            hours=self.user.disconnected_max
        )

    @property
    def is_subject_to_disconnected_termination(self):
        """The related user has disconnected-session-termination enabled."""
        return self.user.disconnected_max > 0

    @property
    def exceeds_disconnected_limit(self):
        """The session's disconnection time exceeds the related user's limit."""
        limit = self.user.disconnected_max * 60 * 60
        if limit <= 0:
            return False

        if self.disconnected_since > limit:
            return True

        return False

    @classmethod
    def is_up_to_date(cls, psy_session):
        """Update (or create) session details from a PSyTricks session dict.

        Parameters
        ----------
        psy_session : dict
            A dict with session details as returned by
            `psytricks.wrapper.ResTricksWrapper.get_sessions()`.

        Returns
        -------
        session, created : diop.models.Session, bool
            A tuple with the first element being the Diop session record and the
            second one being a boolean indicating if the record has been newly
            created (`True`) or if an existing record was updated (`False`).
        """
        uid = psy_session["Uid"]
        session, created = Session.get_or_new(
            uid=uid,
            machine=Machine.from_psy_session(psy_session),
            user=User.from_psy_session(psy_session),
            t_start=psy_session["StartTime"],
        )

        # set the `updated` state without triggering post-save signals or similar:
        Session.objects.filter(uid=uid).update(updated=True)

        return session, created

    @classmethod
    def get_or_new(cls, uid, machine, user, t_start):
        """Get a Session object from the DB or create a new one.

        Query the DB for an existing session object looking for the tuple (UID,
        FQDN, username, session start) or create one if nothing is found.

        Note that the method is intentionally NOT called `get_or_create` to
        lower the risk of mixing it up with Django's built-in model manager one.

        Parameters
        ----------
        uid : int
            The session UID.
        machine : Machine
            The Machine object (or FQDN string).
        user : User
            The user object.
        t_start : datetime
            The session start timestamp.

        Returns
        -------
        (session, created) : (diop.models.Session, bool)
        """
        d_session, created = Session.objects.get_or_create(
            uid=uid,
            machine=machine,
            user=user,
            t_start=make_aware(t_start),
        )
        if created:
            log.success(f"✨🎬 Created NEW session entry: {d_session}")
        else:
            log.debug(f"🎬 Session already known to DB: {d_session} 🆗")
        return d_session, created

    @classmethod
    def last_completed(cls, dg, user):
        """Get the last completed session of a user in the given delivery group.

        Query the DB for all session of the given user in the specified DG and
        return the last one that is already completed / finished.

        Parameters
        ----------
        dg : DeliveryGroup
            The DG to which the session belongs.
        user : User
            The session user.accessgroup_member

        Returns
        -------
        Session
        """
        session = (
            Session.objects.filter(
                user=user, machine__deliverygroup=dg, t_end__isnull=False
            )
            .order_by("-t_end")
            .first()
        )

        return session



from datetime import datetime

from django.contrib.admin.views.decorators import staff_member_required
from django.contrib.auth.decorators import login_required, permission_required
from django.contrib.auth.mixins import LoginRequiredMixin, UserPassesTestMixin
from django.core.exceptions import PermissionDenied
from django.db.models import OuterRef, Subquery
from django.db.models.aggregates import Count, Max
from django.http import Http404, HttpResponse, HttpResponseBadRequest, JsonResponse
from django.shortcuts import redirect, render, get_object_or_404
from django.utils.timezone import make_aware
from django.views.generic import ListView, TemplateView

from loguru import logger as log

from . import tasks
from .models import (
    DeliveryGroup,
    Machine,
    Session,
    SessionProperties,
    Task,
    Issue,
    Notification,
    PpmsBooking,
    Reservation,
)


class MachineList(ListView):
    queryset = {}
    template_name = "diop/machine_list.html"

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)

        machines = Machine.objects.filter(active=True)
        for machine in machines:
            tr_class = ""
            if machine.powerstate == "off":
                tr_class = "table-danger"
            elif machine.maintenance:
                tr_class = "table-warning"
            elif machine.registration != "registered":
                tr_class = "table-info"
            machine.tr_class = tr_class

        context["machines"] = machines
        return context


class MachineListBasic(MachineList):
    queryset = {}
    template_name = "diop/machine_list_basic.html"


class SessionDetails(LoginRequiredMixin, ListView):
    def get(self, request, *args, **kwargs):
        uid = kwargs.get("uid")
        context = {
            "session": Session.objects.get(uid=uid),
            "session_properties": SessionProperties.objects.filter(session=uid),
            "session_notifications": Notification.objects.filter(session=uid),
        }
        return render(request, "diop/session_details.html", context)

    def post(self, request, *args, **kwargs):
        action = request.POST.get("action", None)
        uid = kwargs.get("uid")

        if action is None:
            return HttpResponseBadRequest("Action missing.")
        elif action == "back":
            return redirect("session_list")

        if uid is None:
            return HttpResponseBadRequest("Session UID missing.")

        if action == "terminate":
            log.debug(f"Will request session [{uid}] to be terminated...")
            tasks.terminate_session(uid)
            return redirect("session_details", uid)

        elif action == "disconnect":
            log.debug(f"Will request session [{uid}] to be disconnected...")
            tasks.disconnect_session(uid)
            return redirect("session_details", uid)

        elif action == "message":
            # log.warning(request.POST)
            title = request.POST.get("msg-title", None)
            body = request.POST.get("msg-body", None)
            style = request.POST.get("btn-message-style", "Information")
            if not title or not body:
                return HttpResponseBadRequest("Message title or body missing.")

            log.debug(f"Will send a '{style}' message to session [{uid}]...")
            tasks.send_popup_message(uid, body, title, style)
            return redirect("session_details", uid)

        return HttpResponseBadRequest(f"Unknown action: [{action}]")


class SessionList(LoginRequiredMixin, ListView):
    def get_queryset(self):
        sort = self.request.GET.get("sort", "uid")
        last_props = SessionProperties.objects.filter(session=OuterRef("pk")).order_by(
            "-t_change"
        )
        cur_sessions = (
            Session.objects.filter(t_end=None)
            .annotate(state=Subquery(last_props.values("state")[:1]))
            .annotate(client_address=Subquery(last_props.values("client_address")[:1]))
            .annotate(t_change=Subquery(last_props.values("t_change")[:1]))
            .order_by(sort)
        )

        return cur_sessions


class ReservationDetails(LoginRequiredMixin, ListView):
    queryset = {}
    template_name = "diop/reservation_details.html"

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        id = self.kwargs.get("id")  # NOTE: self.kwargs != kwargs
        context["res_id"] = id
        context["res"] = Reservation.objects.get(id=id)
        context["bookings"] = PpmsBooking.objects.filter(reservation=id).order_by(
            "t_start"
        )
        return context


class OperatorsDash(UserPassesTestMixin, TemplateView):

    template_name = "diop/operators_dash.html"

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        latest = SessionProperties.objects.filter(session=OuterRef("pk")).order_by(
            "-t_change"
        )
        cur_sessions = Session.objects.filter(t_end=None).annotate(
            latest_state=Subquery(latest.values("state")[:1])
        )
        active = cur_sessions.filter(latest_state="active").count()
        disconnected = cur_sessions.exclude(latest_state="active").count()

        context["sessions_active_count"] = active
        context["sessions_disconnected_count"] = disconnected

        context["booking_groups"] = DeliveryGroup.objects.filter(booking_required=True)

        context["reservations_current"] = Reservation.objects.filter(
            t_start__lte=make_aware(datetime.now()),
            t_end__gte=make_aware(datetime.now()),
        ).order_by("user", "-t_end")

        context["reservations_future"] = Reservation.objects.filter(
            t_start__gte=make_aware(datetime.now()),
        ).order_by("user", "-t_end")

        return context

    def test_func(self):
        """Verify the user is a member of the 'Operators' or 'Admins' group.

        Used by the UserPassesTestMixin class.

        Returns
        -------
        bool
        """
        user_groups = self.request.user.groups.all()
        for group in user_groups:
            if group.name in ["Operators", "Admins"]:
                return True
        return False

    def post(self, request, *args, **kwargs):
        action = request.POST.get("action", None)

        if action is None:
            return HttpResponseBadRequest("Action missing.")

        if action == "ppms-pull-users":
            tasks.pull_ppms_users()
            return redirect("operators_dash")
        elif action == "ppms-pull-bookings":
            date = datetime.strptime(request.POST.get("booking-date"), r"%Y-%m-%d")
            days = int(request.POST.get("booking-days"))
            log.warning(f"Requesting bookings for {date} (+ {days-1} days)...")
            tasks.pull_ppms_bookings(date, days)
            return redirect("operators_dash")

        return HttpResponseBadRequest(f"Unknown action: [{action}]")


@staff_member_required
def update_item(request, item):
    referer = request.META.get("HTTP_REFERER", "/")
    if item == "machines":
        okay, fail = tasks.update_machine_status()
    elif item == "sessions":
        okay, fail = tasks.update_session_status()
    else:
        raise Http404("What should we update?")

    return redirect(referer)


def dg_states(request):
    dg_states = DeliveryGroup.objects.values(
        "machine__deliverygroup", "machine__state"
    ).annotate(count=Count("dg_name"))

    # first regroup the results into a dict using the DG name as the key:
    regroup = {}
    for dg_state in dg_states:
        dg = dg_state["machine__deliverygroup"]
        state = dg_state["machine__state"]
        count = dg_state["count"]
        if dg in regroup:
            regroup[dg][state] = count
        else:
            regroup[dg] = {state: count}

    # now use the dict to create a list of dicts:
    state_by_dg = []
    for key, val in regroup.items():
        tmp = {
            "name": key,
            "available": val.get("available", 0),
            "inuse": val.get("inuse", 0),
            "disconnected": val.get("disconnected", 0),
        }
        # log.warning(tmp)
        state_by_dg.append(tmp)

    # log.warning(state_by_dg)
    context = {
        "dg_states": state_by_dg,
    }

    return render(request, "diop/dg_availability.html", context)


@staff_member_required
def status_issues_unresolved(request):
    unresolved = Issue.objects.filter(resolved=False).count()
    if unresolved == 0:
        return HttpResponse("")

    context = {"issues_unresolved_count": unresolved}

    return render(request, "diop/status/issues_unresolved.html", context)


@staff_member_required
def last_update_item(request, item):
    task = f"diop.tasks.update_{item[:-1]}_status"
    log.debug(f"taskname: {task}")

    task_obj = Task.objects.filter(name=task).first

    context = {"item_name": item, "item_label": item.title(), "item_task": task_obj}

    return render(request, "diop/status/update_item.html", context)



"""Management command `pdoc`.

This is required as calling `pdoc` normally results in an error message saying
`Error importing 'module': AppRegistryNotReady: Apps aren't loaded yet.` or
similar.

Example
-------

```bash
SETTINGS_ENV=diop.env diop-manage \
    pdoc \
    -o build/docs \
    -d numpy \
    diop \
    diop_core \
    !diop.migrations \
    !diop.management
```
"""

from pathlib import Path
from django.core.management.base import BaseCommand
from pdoc.__main__ import cli, parser


class Command(BaseCommand):
    """Command class for the `pdoc` management command."""

    def run_from_argv(self, argv: list[str]) -> None:
        """Entry point for `diop-manage`.

        Parameters
        ----------
        argv : list[str]
            The command line argument vector that is passed on by `diop-manage`
            (or `django-admin` for that matter).
        """
        parser.prog = " ".join((Path(argv[0]).name, argv[1]))
        cli(argv[2:])



# Generated by Django 4.2.2 on 2023-06-21 12:35

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("diop", "0021_alter_issue_description_alter_issue_task_and_more"),
    ]

    operations = [
        migrations.AlterField(
            model_name="notification",
            name="body",
            field=models.TextField(verbose_name="The notification body / content."),
        ),
        migrations.AlterField(
            model_name="task",
            name="description",
            field=models.TextField(blank=True, default=""),
            preserve_default=False,
        ),
    ]



# Generated by Django 4.2.2 on 2023-07-11 20:03

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("diop", "0028_deliverygroup_unique_session"),
    ]

    operations = [
        migrations.AlterField(
            model_name="deliverygroup",
            name="access_users",
            field=models.ManyToManyField(
                blank=True,
                help_text="Users that are explicitly granted access to this DG.",
                to="diop.user",
            ),
        ),
        migrations.AlterField(
            model_name="deliverygroup",
            name="host_prefix",
            field=models.CharField(
                blank=True,
                default="",
                help_text="Hostnames starting with this prefix will be considered members of the DG.",
                max_length=32,
                verbose_name="Hostname Prefix",
            ),
        ),
    ]



# Generated by Django 4.2.2 on 2023-07-06 20:55

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("diop", "0024_alter_sitelog_deliverygroup_alter_sitelog_machine_and_more"),
    ]

    operations = [
        migrations.AlterField(
            model_name="user",
            name="department",
            field=models.CharField(
                blank=True,
                help_text='LDAP attribute "department" (may be set manually but LDAP has priority in case the field is present there).',
                max_length=128,
                null=True,
            ),
        ),
        migrations.AlterField(
            model_name="user",
            name="fullname",
            field=models.CharField(
                blank=True,
                help_text='LDAP attribute "displayName".',
                max_length=128,
                null=True,
                verbose_name="Full Name",
            ),
        ),
        migrations.AlterField(
            model_name="user",
            name="groupname",
            field=models.CharField(
                blank=True,
                help_text="Manually set (cannot be derived reliably via LDAP).",
                max_length=128,
                null=True,
                verbose_name="Group Name",
            ),
        ),
        migrations.AlterField(
            model_name="user",
            name="ppms_fullname",
            field=models.CharField(
                blank=True, max_length=128, null=True, verbose_name="PPMS Full Name"
            ),
        ),
        migrations.AlterField(
            model_name="user",
            name="ppms_group",
            field=models.CharField(
                blank=True, max_length=64, null=True, verbose_name="PPMS Group"
            ),
        ),
    ]



"""Diop User model."""

from loguru import logger as log

from django.db import models


class User(models.Model):
    username = models.CharField(max_length=32, primary_key=True)
    fullname = models.CharField(
        max_length=128,
        verbose_name="Full Name",
        help_text='LDAP attribute "displayName".',
        null=True,
        blank=True,
    )
    groupname = models.CharField(
        max_length=128,
        verbose_name="Group Name",
        help_text="Manually set (cannot be derived reliably via LDAP).",
        null=True,
        blank=True,
    )
    department = models.CharField(
        max_length=128,
        help_text=(
            'LDAP attribute "department" (may be set manually but LDAP has '
            "priority in case the field is present there)."
        ),
        null=True,
        blank=True,
    )
    email = models.CharField(max_length=128)
    ppms_group = models.CharField(
        max_length=64,
        verbose_name="PPMS Group",
        null=True,
        blank=True,
    )
    ppms_fullname = models.CharField(
        max_length=128,
        verbose_name="PPMS Full Name",
        null=True,
        blank=True,
    )
    gracetime = models.IntegerField(
        default=0, help_text="Number of days until data expires on the storage."
    )
    disconnected_max = models.IntegerField(
        verbose_name="Max. disconnected hours.",
        default=0,
        help_text=(
            "Hours after which a disconnected session is terminated, negative values "
            "mean no session termination will be enforced."
        ),
    )
    enabled = models.BooleanField(
        default=True,
        help_text="Account is enabled in AD (2nd bit of `userAccountControl` is 0).",
    )
    accessgroup_member = models.BooleanField(
        default=False,
        help_text="User is in the AD-group permitting platform access.",
    )

    def __str__(self):
        """The "local" part of the user's email.

        If the domain part contains something else than the configured site
        domain, that part is added as well.
        """
        # TODO: introduce setting DIOP_DOMAIN setting and adjust logic accordingly!
        mailname = self.email
        try:
            mailname, domain = self.email.split("@")
            if domain.startswith("stud"):
                mailname += " (s)"
        except:
            pass

        return mailname

    @property
    def full_details(self):
        return (
            f"username: {self.username}\n"
            f"fullname: {self.fullname}\n"
            f"groupname: {self.groupname}\n"
            f"department: {self.department}\n"
            f"email: {self.email}\n"
            f"ppms_group: {self.ppms_group}\n"
            f"ppms_fullname: {self.ppms_fullname}\n"
            f"gracetime: {self.gracetime}\n"
            f"disconnected_max: {self.disconnected_max}\n"
            f"enabled: {self.enabled}\n"
            f"accessgroup_member: {self.accessgroup_member}\n"
        )

    @classmethod
    def get_or_new(cls, username: str, defaults: dict = {}):
        r"""Get a user record from the DB or create one if it doesn't exist.

        Note that the method is intentionally NOT called `get_or_create` to
        lower the risk of mixing it up with Django's built-in model manager one.

        Parameters
        ----------
        username : str
            The username. Any leading prefix up to a `\` (backslash) will be
            silently stripped (e.g. `NTDOMAIN\foo` will become `foo`).
        defaults : dict, optional
            A dict with defaults being passed to the `get_or_create()` call.
            The values in the dict will _not_ be used for the record lookup, but
            will be used in case a new one has to be created.

        Returns
        -------
        diop.models.User, bool
            A tuple consisting of the user DB record and a boolean flag
            indicating if the record has been newly created.
        """
        username = username.lower()
        try:
            username = username.split("\\")[1]
        except:  # ruff: noqa: E722
            pass  # if the split fails we're using the string as-is

        log.debug(f"👩‍🔬 Checking for user [{username}]...")

        d_user, created = cls.objects.get_or_create(
            username=username,
            defaults=defaults,
        )

        if created:
            log.success(f"👩‍🔬 Created NEW user entry: [{d_user.username}] ✨")
        else:
            log.debug(f"👩‍🔬 User already known: [{d_user.username}] 🆗")

        return d_user, created

    @classmethod
    def is_up_to_date_from_ad(cls, user_details, accessgroup_member):
        """Create or update a user from an AD query response.

        Parameters
        ----------
        user_details : Bix
            A Box object with user details as returned by
            `diop.adapters.activedirectory.ActiveDirectory.user_details_from_dn`.
        accessgroup_member : bool
            A flag indicating whether the user is a member of the access group.

        Returns
        -------
        (created, updated) : (bool, bool)
            A tuple indiciating if the user record has been created or updated.
        """
        return cls.is_up_to_date(
            username=user_details.username,
            email=user_details.email,
            fullname=user_details.display_name,
            department=user_details.department,
            enabled=user_details.enabled,
            accessgroup_member=accessgroup_member,
        )

    @classmethod
    def is_up_to_date(
        cls, username, email, fullname, department, enabled, accessgroup_member
    ):
        """Create or get a user record, ensuring all attributes are up-to-date.

        Fetch an existing user record (identified via the PK `username`) or
        create a new record using the given details. In case an existing record
        is found, it is made sure the non-PK attributes match the ones given as
        parameters to this method.

        Parameters
        ----------
        username : str
            The username (=primary key).
        email : str
            The user's email address.
        fullname : str
            The full name.
        department : str
            The department
        enabled : bool
            A flag indicating if the user account is enabled in ActiveDirectory.
        accessgroup_member : bool
            A flag indicating whether the user is a member of the access group.

        Returns
        -------
        (created, updated) : (bool, bool)
            A tuple indiciating if the user record has been created or updated.
        """
        updated = False
        defaults = {
            "email": email,
            "fullname": fullname,
            "department": department,
            "enabled": enabled,
            "accessgroup_member": accessgroup_member,
        }
        d_user, created = cls.get_or_new(username=username, defaults=defaults)

        if created or (
            d_user.email == email
            and d_user.fullname == fullname
            and d_user.department == department
            and d_user.enabled == enabled
            and d_user.accessgroup_member
        ):
            log.debug(f"👩‍🔬 No further user changes: [{d_user.username}] 🆗")
            return created, updated

        updated = True
        d_user.email = email
        d_user.fullname = fullname
        d_user.department = department
        d_user.enabled = enabled
        # TODO: why us `True` hardcoded here for the accessgroup_member, is this
        # a bug? If not, explain it here (and in the docstring)!
        d_user.accessgroup_member = True
        d_user.save()

        return created, updated

    @classmethod
    def from_psy_session(cls, psy_session: dict):
        """Get a DIOP user object from a PSyTricks session dict.

        Parameters
        ----------
        psy_session : dict
            Session infos as returned by psytricks.ResTricksWrapper.get_sessions().

        Returns
        -------
        diop.models.User
            The DB record corresponding to the session user.
        """
        username = psy_session["UserName"]
        upn = psy_session["UserUPN"]

        d_user, _ = cls.get_or_new(username, defaults={"email": upn})

        return d_user



"""Diop Reservation model."""

from datetime import timedelta
from datetimerange import DateTimeRange

from django.db import models
from django.utils import timezone

from ..utils import naive_minutes
from .user import User
from .deliverygroup import DeliveryGroup


class Reservation(models.Model):
    class Meta:
        constraints = [
            models.UniqueConstraint(
                fields=["user", "deliverygroup", "t_start", "t_end"],
                name="unique_reservation",
            )
        ]

    user = models.ForeignKey(User, on_delete=models.CASCADE)
    deliverygroup = models.ForeignKey(
        DeliveryGroup,
        on_delete=models.CASCADE,
        verbose_name="Delivery Group",
        help_text="Delivery Group the reservation is for.",
    )
    t_start = models.DateTimeField("Reservation Start", null=False, blank=False)
    t_end = models.DateTimeField("Reservation End", null=False, blank=False)

    def __str__(self):
        return (
            f"({self.pk}) {self.user.username}@{self.deliverygroup} "
            f"[{naive_minutes(self.t_start)} -- {naive_minutes(self.t_end)}]"
        )

    @property
    def t_range(self):
        return DateTimeRange(self.t_start, self.t_end)

    @property
    def till_end(self):
        """The number of seconds until the end of the reservation.

        Returns
        -------
        float
            The time (in seconds) that is left until the reservation ends, will
            be negative if the reservation's end is in the past.
        """
        return (self.t_end - timezone.now()).total_seconds()

    @property
    def since_start(self):
        """The number of seconds since the start of the reservation.

        Returns
        -------
        float
            The time (in seconds) that has passed since the reservation's start,
            will be negative if the start time is in the future.
        """
        return (timezone.now() - self.t_start).total_seconds()

    @classmethod
    def running(cls, dg_name):
        """The currently active ("running") reservations for the given DG.

        Parameters
        ----------
        dg_name : str
            The name of the delivery group to check reservations for.

        Returns
        -------
        QuerySet
            All reservations whose starttime is in the past and endtime is in
            the future (both also including "now").
        """
        now = timezone.now()
        return cls.objects.filter(
            deliverygroup__dg_name=dg_name, t_start__lte=now, t_end__gte=now
        )

    @classmethod
    def running_usernames(cls, dg_name):
        """A list of usernames having currently active ("running") reservations.

        Parameters
        ----------
        dg_name : str
            The name of the delivery group to check reservations for.

        Returns
        -------
        QuerySet
            All usernames having a currently active reservation as given by
            `Reservation.running()`.
        """
        running = cls.running(dg_name)
        return running.values_list("user__username", flat=True)

    @classmethod
    def ending_in_less_than(cls, minutes: int, dg_name: str, username: str = ""):
        """Running reservations ending in less than the given delta from now.

        Parameters
        ----------
        minutes : int
            The amount of minutes to use as the limit for checking when the
            reservations will expire.
        dg_name : str
            The name of the delivery group to check reservations for.
        username : str, optional
            An optional username to filter the reservations for, by default ""
            which will result in no filtering being done.

        Returns
        -------
        QuerySet
            All running reservations belonging to the given DG that will expire
            in less than the given number of minutes, optionally restricted by
            the username (if given).
        """
        running = cls.running(dg_name)
        ending = running.filter(
            t_end__lte=(timezone.now() + timedelta(minutes=minutes))
        )
        if not username:
            return ending

        return ending.filter(user__username=username)



# Generated by Django 4.1.7 on 2023-05-22 10:10

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("diop", "0011_task_issue"),
    ]

    operations = [
        migrations.AddField(
            model_name="machine",
            name="active",
            field=models.BooleanField(
                default=True,
                help_text="Indicating if the machine is active or has been retired. Entries having this set to False will be ignored by update task completeness checks.",
                verbose_name="Active",
            ),
        ),
        migrations.AddField(
            model_name="machine",
            name="updated",
            field=models.BooleanField(
                default=True,
                help_text="Flag tracking if the machine status is up-to-date. Has to be set to False by every function updating an entry and only re-set to True once the update has completed successfully.",
                verbose_name="Updated",
            ),
        ),
    ]



from django.apps import AppConfig


class DiopConfig(AppConfig):
    name = "diop"
    verbose_name = "DIOP"

    def ready(self):
        # implicitly connect signal handlers decorated with @receiver.
        from . import signals  # ruff: noqa: F401 (imported but unused)



# Generated by Django 4.2.2 on 2023-06-13 20:20

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):
    dependencies = [
        ("diop", "0016_notification"),
    ]

    operations = [
        migrations.AlterField(
            model_name="machine",
            name="deliverygroup",
            field=models.ForeignKey(
                help_text="Delivery Group the machine belongs to.",
                on_delete=django.db.models.deletion.CASCADE,
                to="diop.deliverygroup",
                verbose_name="Delivery Group",
            ),
        ),
        migrations.AlterField(
            model_name="machine",
            name="state",
            field=models.CharField(
                help_text="Current machine status.",
                max_length=16,
                verbose_name="Machine State",
            ),
        ),
    ]



"""Diop exception classes."""

import logging

log = logging.getLogger(__name__)


class UpdateTaskError(Exception):
    def __init__(self, message, details, okay, fail):
        self.message = message
        self.details = details
        self.okay = okay
        self.fail = fail

    def __str__(self):
        details = "\n".join(str(self.details))
        return (
            f"{self.message}\n"
            f"OKAY={self.okay}\n"
            f"FAIL={self.fail}\n"
            f"FAILED DETAILS:\n{details}"
        )


class UpdateMachinesError(UpdateTaskError):
    pass


class UpdateSessionsError(UpdateTaskError):
    pass



# Generated by Django 4.2.2 on 2023-07-07 21:57

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("diop", "0026_remove_user_active_user_enabled"),
    ]

    operations = [
        migrations.DeleteModel(
            name="SiteConfig",
        ),
    ]



# Generated by Django 4.1.7 on 2023-05-15 20:29

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("diop", "0009_alter_session_user"),
    ]

    operations = [
        migrations.AddField(
            model_name="machine",
            name="agent",
            field=models.CharField(
                max_length=32, null=True, verbose_name="Agent Version"
            ),
        ),
    ]



# Generated by Django 4.1.7 on 2023-04-14 14:04

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):
    dependencies = [
        ("diop", "0004_deliverygroup"),
    ]

    operations = [
        migrations.AddField(
            model_name="machine",
            name="deliverygroup",
            field=models.ForeignKey(
                default="Dinner",
                on_delete=django.db.models.deletion.CASCADE,
                to="diop.deliverygroup",
            ),
            preserve_default=False,
        ),
    ]



"""Diop Notification model."""

from typing import Callable

from django.db import models
from django.utils.timezone import now

from loguru import logger as log

from ..utils import send_email
from .user import User
from .session import Session
from .message_template import MessageTemplate


class Notification(models.Model):
    class NotificationType(models.TextChoices):
        POP_UP = "P"
        EMAIL = "E"

    user = models.ForeignKey(
        User,
        on_delete=models.CASCADE,
        help_text="User which the notification is addressed to.",
        null=False,
        blank=False,
    )
    session = models.ForeignKey(
        Session,
        on_delete=models.CASCADE,
        help_text="Session the notification relates to.",
        null=True,
        blank=True,
    )
    method = models.CharField(
        max_length=1,
        choices=NotificationType.choices,
        default=NotificationType.POP_UP,
        verbose_name="Delivery Method",
        help_text="How the notification should be delievered to the user.",
    )
    cc_adm = models.BooleanField(
        verbose_name="Cc Operators",
        help_text="Flag if admins should be notified as well.",
    )
    subject = models.CharField(
        max_length=256,
        null=False,
        blank=False,
        verbose_name="The notification subject / title.",
    )
    body = models.TextField(
        null=False,
        blank=False,
        verbose_name="The notification body / content.",
    )
    t_created = models.DateTimeField(
        verbose_name="Creation Time",
        help_text="Notification creation timestamp.",
        null=False,
        default=now,
    )
    t_sent = models.DateTimeField(
        verbose_name="Sent Time",
        help_text="Notification delivery timestamp (empty if not yet delieverd).",
        null=True,
    )
    creator = models.CharField(
        verbose_name="Creator",
        max_length=64,
        help_text="Creator / origin of the notification.",
        blank=True,
    )

    def __str__(self):
        return f"{self.user.email} [{self.subject}]"

    @classmethod
    def new_popup(cls, send_popup: Callable, uid: int, template: str, context: dict):
        """Send a pop-up notification unless it has already been sent before.

        Parameters
        ----------
        send_popup : Callable
            A callable taking care of delivering the message to a session.
        uid : int
            The session UID.
        template : str
            The name of the template to use for the message.
        context : dict
            The context dict required to render the template.
        """
        try:
            msg = MessageTemplate.render_message(template, context)
        except Exception as ex:
            log.error(f"Rendering message template [{template}] failed: {ex}")
            return False

        session = Session.objects.get(pk=uid)
        if (
            cls.objects.exclude(t_sent=None)
            .filter(
                user=session.user,
                session=session,
                method=cls.NotificationType.POP_UP,
                subject=msg.subject_raw,
                body=msg.body_raw,
            )
            .exists()
        ):
            log.trace("Message has been delivered before, not re-sending.")
            return False

        notification = cls(
            user=session.user,
            session=session,
            method="P",
            cc_adm=False,
            subject=msg.subject_raw,
            body=msg.body_raw,
        )
        notification.save()
        machine = session.machine.fqdn
        log.debug(f"🖥🎬📢 <send_message> ⏩ [{uid}]@[{machine}]")
        sent = send_popup(
            machine=machine,
            body=msg.body,
            subject=msg.subject,
            category=msg.category,
        )
        if sent:
            notification.t_sent = now()
            notification.save()
            return True

        # should we raise an exception instead, to distinguish between messages
        # that have been sent before (no-re-delivery) and sending errors?
        return False

    @classmethod
    def new_email(
        cls,
        user: User,
        template: str,
        context: dict,
        support: str = "",
        session: Session = None,
    ):
        """Send an email notification unless it has already been sent before.

        If a message is actually sent, a new notification will be recorded in
        the DB, containing all relevant properties of the message. To ensure
        this record is independent of external circumstances, the "raw" versions
        of the message subject and body will be used, i.e. the strings that were
        rendered from the given template with all Django template filters being
        stripped. This ensures messages will not be re-sent if e.g. a
        `naturaltime` filter produces a different result to when the
        notification has been initially recorded in the DB.

        Attempting to send an identical message multiple times with different
        values for the `support` parameter (e.g. once with `support=""` and
        once with `support="cc"` or `support="bcc"`) will deliver the message
        twice, once only to the user and once having the admins in Cc.

        Parameters
        ----------
        user : User
            The Diop user related to this email.
        template : str
            The name of the template to use for the message.
        context : dict
            The context dict required to render the template.
        support : str, optional
            If set to `cc` or `bcc` the address set in `DIOP_SUPPORT_CONTACT`
            will be added as a respective recipient. By default empty.
        session : Session, optional
            An optional session object that is related to this email (will be
            used to link the created notification in the DB accordingly). By
            default `None`.

        Raises
        ------
        SMTPException
            May be raised by the `send_email()` call.

        Returns
        -------
        bool
            True in case the email was sent, False (e.g. rendering failed, or
            the message has already been sent before) otherwise.
        """
        try:
            msg = MessageTemplate.render_message(template, context)
        except Exception as ex:
            log.error(f"Rendering message template [{template}] failed: {ex}")
            return False

        if (
            cls.objects.exclude(t_sent=None)
            .filter(
                user=user,
                session=session,
                method=cls.NotificationType.EMAIL,
                cc_adm=True if support in ["cc", "bcc"] else False,
                subject=msg.subject_raw,
                body=msg.body_raw,
            )
            .exists()
        ):
            log.trace("Message has been delivered before, not re-sending.")
            return False

        notification = cls(
            user=user,
            session=session,
            method=cls.NotificationType.EMAIL,
            cc_adm=True if support in ["cc", "bcc"] else False,
            subject=msg.subject_raw,
            body=msg.body_raw,
        )
        notification.save()

        send_email(subject=msg.subject, body=msg.body, to=user.email, support=support)

        notification.t_sent = now()
        notification.save()
        return True



from django.conf import settings
from django.db.models.signals import post_save
from django.dispatch import receiver

from loguru import logger as log

from .models import (
    SiteLog,
    User,
    Reservation,
    Notification,
    Session,
    SessionProperties,
    Machine,
)


@receiver(post_save, sender=User)
def user_post_save_handler(sender, instance, created, **kwargs):
    log.trace(f"Received user post-save signal (created={created}): {instance}")

    if created:
        name = "User added"
    else:
        name = "User modified"

    record = SiteLog(name=name, user=instance, details=instance.full_details)
    record.save()

    # update the disconnected_max value with the default in case it's 0:
    User.objects.filter(pk=instance.pk, disconnected_max=0).update(
        disconnected_max=settings.DIOP_DISCONNECTED_MAX
    )


# FIXME: add a signal handler for deleting reservations!!


@receiver(post_save, sender=Reservation)
def reservation_post_save_handler(sender, instance, created, **kwargs):
    log.trace(f"Received reservation post-save signal (created={created}): {instance}")

    if created:
        name = "Reservation added"
    else:
        name = "Reservation modified"

    record = SiteLog(
        name=name,
        reservation=instance,
        user=instance.user,
        deliverygroup=instance.deliverygroup,
        details=str(instance),
    )
    record.save()


@receiver(post_save, sender=Notification)
def notification_post_save_handler(sender, instance, created, **kwargs):
    notification = instance
    if not notification.t_sent:
        log.debug(f"Notification created but not sent yet: {notification}")
        return

    log.trace(f"Notification post-save signal (created={created}): {notification}")

    details = (
        f"{notification}\n\n"
        f"method=[{notification.method}]\n"
        f"cc_adm=[{notification.cc_adm}]\n"
        f"t_created=[{notification.t_created}]\n"
        f"t_sent=[{notification.t_sent}]\n"
        f"creator=[{notification.creator}]\n"
        f"subject=[{notification.subject}]\n"
        f"body=[{notification.body}]\n"
    )

    # email notifications don't necessarily have a session (and therefore machine, ...)
    machine = deliverygroup = None
    if notification.session:
        machine = notification.session.machine
        deliverygroup = notification.session.machine.deliverygroup

    record = SiteLog(
        name="Notification sent",
        parameters=f"subject=[{notification.subject}]",
        notification=notification,
        session=notification.session,
        user=notification.user,
        machine=machine,
        deliverygroup=deliverygroup,
        details=details,
    )
    record.save()


@receiver(post_save, sender=Machine)
def machine_post_save_handler(sender, instance, created, **kwargs):
    machine = instance
    log.trace(f"Machine post-save signal (created={created}): {machine}")

    details = (
        f"{machine}\n"
        f"state=[{machine.state}]\n"
        f"powerstate=[{machine.powerstate}]\n"
        f"maintenance=[{machine.maintenance}]\n"
        f"registration=[{machine.registration}]\n"
        f"agent=[{machine.agent}]\n"
        f"active=[{machine.active}]\n"
        f"req_maintenance=[{machine.req_maintenance}]\n"
    )

    maint_short = "maint" if machine.maintenance else "no-maint"
    reg_short = "reg" if machine.registration == "registered" else "unreg"
    parameters = f"{machine.powerstate} | {machine.state} | {reg_short} | {maint_short}"

    log_name = "Machine status changed"
    if created:
        log_name = "Machine added"

    record = SiteLog(
        name=log_name,
        machine=machine,
        deliverygroup=machine.deliverygroup,
        parameters=parameters,
        details=details,
    )
    record.save()


@receiver(post_save, sender=Session)
def session_post_save_handler(sender, instance, created, **kwargs):
    session = instance
    log.trace(f"Session post-save signal (created={created}): {session}")

    details = (
        f"{session}\n"
        f"machine=[{session.machine}]\n"
        f"user=[{session.user}]\n"
        f"t_start=[{session.t_start}]\n"
        f"t_end=[{session.t_end}]\n"
    )

    log_name = "Session ended"
    if created:
        log_name = "Session started"

    record = SiteLog(
        name=log_name,
        session=session,
        user=session.user,
        machine=session.machine,
        deliverygroup=session.machine.deliverygroup,
        parameters=f"t_start=[{session.t_start}]",
        details=details,
    )
    record.save()


@receiver(post_save, sender=SessionProperties)
def sessionproperties_post_save_handler(sender, instance, created, **kwargs):
    props = instance
    log.trace(f"SessionProperties post-save signal (created={created}): {props}")

    details = (
        f"{props}\n"
        f"state=[{props.state}]\n"
        f"t_change=[{props.t_change}]\n"
        f"client_address=[{props.client_address}]\n"
        f"client_name=[{props.client_name}]\n"
        f"client_version=[{props.client_version}]\n"
    )

    record = SiteLog(
        name="Session Properties changed",
        parameters=f"state=[{props.state}]",
        session=props.session,
        user=props.session.user,
        machine=props.session.machine,
        deliverygroup=props.session.machine.deliverygroup,
        details=details,
    )
    record.save()



# Generated by Django 4.2.2 on 2023-06-17 17:38

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):
    dependencies = [
        ("diop", "0019_user_disconnected_max_alter_user_accessgroup_member_and_more"),
    ]

    operations = [
        migrations.CreateModel(
            name="Reservation",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("t_start", models.DateTimeField(verbose_name="Reservation Start")),
                ("t_end", models.DateTimeField(verbose_name="Reservation End")),
                (
                    "deliverygroup",
                    models.ForeignKey(
                        help_text="Delivery Group the reservation is for.",
                        on_delete=django.db.models.deletion.CASCADE,
                        to="diop.deliverygroup",
                        verbose_name="Delivery Group",
                    ),
                ),
                (
                    "user",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE, to="diop.user"
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="PpmsBooking",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "username",
                    models.CharField(max_length=32, verbose_name="PPMS Username"),
                ),
                ("system_id", models.IntegerField(verbose_name="PPMS System ID")),
                ("t_start", models.DateTimeField(verbose_name="Booking Start")),
                ("t_end", models.DateTimeField(verbose_name="Booking End")),
                (
                    "ppms_session",
                    models.CharField(
                        blank=True,
                        max_length=8,
                        null=True,
                        verbose_name="PPMS Session ID",
                    ),
                ),
                (
                    "sync_state",
                    models.CharField(
                        choices=[("D", "Deleted"), ("N", "New"), ("K", "Known")],
                        default="N",
                        max_length=1,
                    ),
                ),
                (
                    "reservation",
                    models.ForeignKey(
                        blank=True,
                        help_text="Reservation to which this booking belongs.",
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        to="diop.reservation",
                    ),
                ),
            ],
            options={
                "verbose_name": "PPMS Booking",
            },
        ),
        migrations.AddConstraint(
            model_name="reservation",
            constraint=models.UniqueConstraint(
                fields=("user", "deliverygroup", "t_start", "t_end"),
                name="unique_reservation",
            ),
        ),
        migrations.AddConstraint(
            model_name="ppmsbooking",
            constraint=models.UniqueConstraint(
                fields=("username", "system_id", "t_start", "t_end"),
                name="unique_booking",
            ),
        ),
    ]



# Generated by Django 4.2.2 on 2023-06-16 18:53

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("diop", "0018_machine_req_maintenance"),
    ]

    operations = [
        migrations.AddField(
            model_name="user",
            name="disconnected_max",
            field=models.IntegerField(
                default=0,
                help_text="Hours after which a disconnected session is terminated, negative values mean no session termination will be enforced.",
                verbose_name="Max. disconnected hours.",
            ),
        ),
        migrations.AlterField(
            model_name="user",
            name="accessgroup_member",
            field=models.BooleanField(
                default=False,
                help_text="User is in the AD-group permitting platform access.",
            ),
        ),
        migrations.AlterField(
            model_name="user",
            name="fullname",
            field=models.CharField(max_length=128, verbose_name="Full Name"),
        ),
        migrations.AlterField(
            model_name="user",
            name="gracetime",
            field=models.IntegerField(
                default=0, help_text="Number of days until data expires on the storage."
            ),
        ),
        migrations.AlterField(
            model_name="user",
            name="groupname",
            field=models.CharField(max_length=128, verbose_name="Group Name"),
        ),
        migrations.AlterField(
            model_name="user",
            name="ppms_fullname",
            field=models.CharField(max_length=128, verbose_name="PPMS Full Name"),
        ),
        migrations.AlterField(
            model_name="user",
            name="ppms_group",
            field=models.CharField(max_length=64, verbose_name="PPMS Group"),
        ),
    ]



# Generated by Django 4.1.7 on 2023-05-21 21:17

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):
    dependencies = [
        ("diop", "0010_machine_agent"),
    ]

    operations = [
        migrations.CreateModel(
            name="Task",
            fields=[
                (
                    "name",
                    models.CharField(
                        max_length=128,
                        primary_key=True,
                        serialize=False,
                        verbose_name="Task Name",
                    ),
                ),
                ("t_start", models.DateTimeField(verbose_name="Start Time")),
                (
                    "t_end",
                    models.DateTimeField(
                        blank=True,
                        help_text="Empty if task is still running or did not succeed.",
                        null=True,
                        verbose_name="End Time",
                    ),
                ),
                (
                    "unfinished",
                    models.SmallIntegerField(
                        default=0,
                        help_text="Counter on previous unfinished runs when next one tries to start.",
                    ),
                ),
                (
                    "failed",
                    models.SmallIntegerField(
                        default=0, help_text="Counter on previously failed attempts."
                    ),
                ),
                (
                    "description",
                    models.CharField(blank=True, max_length=128, null=True),
                ),
            ],
        ),
        migrations.CreateModel(
            name="Issue",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "severity",
                    models.SmallIntegerField(
                        default=0, help_text="Higher means more critical."
                    ),
                ),
                (
                    "timestamp",
                    models.DateTimeField(help_text="Time the issue occurred."),
                ),
                (
                    "resolved",
                    models.BooleanField(
                        help_text="Flag indicating the issue is resolved."
                    ),
                ),
                ("description", models.CharField(max_length=512)),
                (
                    "task",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        to="diop.task",
                        verbose_name="Task causing the issue (if applicable).",
                    ),
                ),
                (
                    "user",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        to="diop.user",
                        verbose_name="User related to the issue (if applicable).",
                    ),
                ),
            ],
        ),
    ]



# Generated by Django 4.2.2 on 2023-07-11 19:55

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("diop", "0027_delete_siteconfig"),
    ]

    operations = [
        migrations.AddField(
            model_name="deliverygroup",
            name="unique_session",
            field=models.BooleanField(
                default=True,
                help_text="Allow only one session per user across *ALL* groups having this flag enabled (enforced via a housekeeping task).",
            ),
        ),
    ]



# Generated by Django 4.1.7 on 2023-04-17 15:43

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):
    dependencies = [
        ("diop", "0008_session_sessionproperties"),
    ]

    operations = [
        migrations.AlterField(
            model_name="session",
            name="user",
            field=models.ForeignKey(
                blank=True,
                help_text="User associated to the session, can be empty!",
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                to="diop.user",
                verbose_name="User",
            ),
        ),
    ]



"""Diop SessionProperties model."""

from datetime import datetime

from loguru import logger as log

from django.db import models
from django.utils.timezone import make_aware

from .session import Session


class SessionProperties(models.Model):
    session = models.ForeignKey(
        Session, on_delete=models.CASCADE, verbose_name="Session UID"
    )
    t_change = models.DateTimeField(
        "State Change Time",
        help_text="Citrix:SessionStateChangeTime",
    )
    state = models.CharField("Session State", max_length=16)
    client_address = models.CharField(
        "Client IP",
        max_length=46,
        help_text="IP address of the client currently (or last) connected.",
    )
    client_name = models.CharField(
        "Client Name",
        max_length=64,
        help_text="The name of the client currently (or last) connected.",
    )
    client_version = models.CharField(
        "Client Version",
        max_length=16,
        help_text="The version of the client currently (or last) connected.",
    )

    class Meta:
        verbose_name = "Session Properties"
        verbose_name_plural = "Session Properties"
        constraints: [
            models.UniqueConstraint(
                fields=["t_change", "state"],
                name="unique_tchange_state_combination",
            )
        ]

    def __str__(self):
        return f"({self.session}) {str(self.t_change)} [{self.state}]"

    @classmethod
    def get_or_new(
        cls,
        session: Session,
        t_change: datetime,
        state: str,
        defaults: dict = {},
    ):
        """Get a SessionProperties object from the DB or create a new one.

        Query the DB for an existing session object looking for the tuple
        (session, t_change, state) or create one if nothing is found.

        Note that the method is intentionally NOT called `get_or_create` to
        lower the risk of mixing it up with Django's built-in model manager one.

        Parameters
        ----------
        session: Session
            The Session to which this object belongs.
        t_change: datetime
        state: str

        Returns
        -------
        (properties, created) : (diop.models.SessionProperties, bool)
        """
        props, created = SessionProperties.objects.get_or_create(
            session=session,
            t_change=make_aware(t_change),
            state=state,
            defaults=defaults,
        )

        # only if the session properties entry was newly created we need to fill
        # in the details, otherwise we can simply return here:
        if created:
            log.success(f"✨🎬🏷 Created NEW session properties: {props}")
        else:
            log.debug(f"🎬🏷 Session properties known, not updating: {props} 🆗")

        return props, created



from django.contrib import admin

from loguru import logger as log

from .models import (
    User,
    Machine,
    DeliveryGroup,
    Session,
    SessionProperties,
    Task,
    Issue,
    Notification,
    Reservation,
    PpmsBooking,
    SiteLog,
    MessageTemplate,
)

# globally disable "delete selected" action:
admin.site.disable_action("delete_selected")


class SessionInline(admin.TabularInline):
    model = Session
    readonly_fields = ("uid", "machine", "user", "t_start")
    exclude = ["t_end"]
    show_change_link = True
    can_delete = False
    extra = 0


class SessionPropertiesInline(admin.TabularInline):
    model = SessionProperties
    readonly_fields = (
        "t_change",
        "state",
        "client_address",
        "client_version",
    )
    exclude = ["client_name"]
    show_change_link = True
    can_delete = False
    extra = 0


@admin.register(DeliveryGroup)
class DeliveryGroupAdmin(admin.ModelAdmin):
    list_display = ("dg_name", "booking_required", "unique_session", "host_prefix")


@admin.register(Machine)
class MachineAdmin(admin.ModelAdmin):
    list_display = (
        "hostname",
        "deliverygroup",
        "maintenance",
        "state",
        "powerstate",
        "active",
        "updated",
        "req_maintenance",
    )
    actions = [
        "request_maintenance",
    ]
    list_filter = [
        "deliverygroup",
        "maintenance",
        "state",
        "active",
        "updated",
        "req_maintenance",
    ]
    readonly_fields = [
        "fqdn",
        "hostname",
        "deliverygroup",
        "registration",
        "maintenance",
        "agent",
        "state",
        "powerstate",
        "updated",
    ]
    search_fields = ("fqdn",)

    @admin.action(description="Toggle 'Request Maintenance'")
    def request_maintenance(self, request, queryset):
        for machine in queryset:
            log.info(f"Toggling 'Request maintenance' for [{machine}]")
            machine.req_maintenance = not (machine.req_maintenance)
            machine.save()


@admin.register(User)
class UserAdmin(admin.ModelAdmin):
    list_display = (
        "username",
        "fullname",
        "ppms_fullname",
        "email",
        "groupname",
        "ppms_group",
        "department",
        "gracetime",
        "disconnected_max",
        "enabled",
        "accessgroup_member",
    )
    list_filter = [
        "enabled",
        "accessgroup_member",
        "gracetime",
        "disconnected_max",
        "department",
        "groupname",
        "ppms_group",
    ]
    search_fields = (
        "username",
        "fullname",
        "ppms_fullname",
        "email",
        "groupname",
        "ppms_group",
        "department",
    )
    readonly_fields = [
        "username",
        "email",
        "fullname",
        # "ppms_fullname",  # FIXME: should this be read-only...?
        # "ppms_group",  # FIXME: should this be read-only...?
        "accessgroup_member",
        "enabled",
    ]
    # TODO: reactivate inlines once the django-edit-bug is fixed
    # inlines = [SessionInline]


@admin.register(Session)
class SessionAdmin(admin.ModelAdmin):
    list_display = ("uid", "machine", "user", "t_start", "t_end")
    list_filter = ["t_end", "user", "machine"]
    readonly_fields = [
        "uid",
        "machine",
        "user",
        "t_start",
        "t_end",
        "updated",
    ]
    # TODO: reactivate inlines once the django-edit-bug is fixed
    # inlines = [SessionPropertiesInline]


@admin.register(SessionProperties)
class SessionPropertiesAdmin(admin.ModelAdmin):
    list_display = ("session", "t_change", "state", "client_address")
    list_filter = ["session", "state"]
    readonly_fields = [
        "session",
        "t_change",
        "state",
        "client_address",
        "client_name",
        "client_version",
    ]


@admin.register(Task)
class TaskAdmin(admin.ModelAdmin):
    list_display = (
        "name",
        "t_start",
        "t_end",
        "unfinished",
        "failed",
        "description",
    )
    list_filter = ["name", "failed"]


@admin.register(Issue)
class IssueAdmin(admin.ModelAdmin):
    list_display = (
        "severity",
        "timestamp",
        "resolved",
        "task",
        "user",
        "description",
    )
    list_filter = ["resolved", "severity", "task", "user"]
    actions = [
        "mark_resolved",
    ]
    search_fields = ("description",)

    @admin.action(description="Mark as resolved")
    def mark_resolved(self, request, queryset):
        for issue in queryset:
            log.info(f"Marking issue as resolved: {issue}")
            issue.resolved = True
            issue.save()


@admin.register(Notification)
class NotificationAdmin(admin.ModelAdmin):
    list_display = (
        "user",
        "session",
        "method",
        "cc_adm",
        "subject",
        "t_created",
        "t_sent",
        "creator",
    )
    list_filter = [
        "t_sent",
        "method",
        "cc_adm",
        "user",
    ]
    search_fields = (
        "subject",
        "body",
    )
    readonly_fields = [
        "user",
        "session",
        "method",
        "cc_adm",
        "subject",
        "body",
        "t_created",
        "t_sent",
        "creator",
    ]


@admin.register(Reservation)
class ReservationAdmin(admin.ModelAdmin):
    list_display = (
        "id",
        "user",
        "deliverygroup",
        "t_start",
        "t_end",
    )
    list_filter = [
        "t_start",
        "deliverygroup",
        "user",
    ]
    readonly_fields = [
        "user",
        "deliverygroup",
        "t_start",
        "t_end",
    ]


@admin.register(PpmsBooking)
class PpmsBookingAdmin(admin.ModelAdmin):
    list_display = (
        "id",
        "username",
        "system_id",
        "t_start",
        "t_end",
        "ppms_session",
        "sync_state",
        "reservation",
    )
    list_filter = [
        "t_start",
        "sync_state",
        "username",
    ]
    readonly_fields = [
        "username",
        "system_id",
        "t_start",
        "t_end",
        "ppms_session",
        "sync_state",
        "reservation",
    ]


@admin.register(SiteLog)
class SiteLogAdmin(admin.ModelAdmin):
    list_display = (
        "name",
        "parameters",
        "time",
        "user",
        "deliverygroup",
        "machine",
        "session",
        "task",
        "notification",
        "reservation",
    )
    list_filter = [
        "name",
        "time",
        "deliverygroup",
        "machine",
        "user",
    ]
    readonly_fields = [
        "name",
        "parameters",
        "time",
        "details",
        "user",
        "deliverygroup",
        "machine",
        "session",
        "task",
        "notification",
        "reservation",
    ]


@admin.register(MessageTemplate)
class MessageTemplateAdmin(admin.ModelAdmin):
    list_display = ("name", "subject")



# Generated by Django 4.1.7 on 2023-05-22 12:04

from django.db import migrations, models
import django.utils.timezone


class Migration(migrations.Migration):
    dependencies = [
        ("diop", "0012_machine_active_machine_updated"),
    ]

    operations = [
        migrations.AlterField(
            model_name="issue",
            name="description",
            field=models.CharField(max_length=4096),
        ),
        migrations.AlterField(
            model_name="issue",
            name="resolved",
            field=models.BooleanField(
                default=False, help_text="Flag indicating the issue is resolved."
            ),
        ),
        migrations.AlterField(
            model_name="issue",
            name="severity",
            field=models.SmallIntegerField(
                choices=[
                    (0, "Low"),
                    (1, "Medium"),
                    (2, "High"),
                    (3, "Critical"),
                    (4, "Meltdown"),
                ],
                default=0,
                help_text="Values from 0 to 4, higher means more critical.",
            ),
        ),
        migrations.AlterField(
            model_name="issue",
            name="timestamp",
            field=models.DateTimeField(
                auto_now_add=True, help_text="Time the issue occurred."
            ),
        ),
        migrations.AlterField(
            model_name="task",
            name="t_start",
            field=models.DateTimeField(
                default=django.utils.timezone.now, verbose_name="Start Time"
            ),
        ),
    ]



"""Diop decorators."""

import logging
from datetime import datetime
from functools import wraps, partial
import traceback

from django.utils.timezone import make_aware

from .exceptions import UpdateMachinesError, UpdateSessionsError
from .models import Task, Issue

log = logging.getLogger(__name__)


def status_update_task(func):
    """Register a task start in the DB, execute it and check the results.

    This decorator will update the timestamp for a given "update" task start (or
    create a new one) using `func`'s name as identifier, then execute the given
    task function and watch out for exceptions.

    *WARNING*: decorated functions will have any exception "silenced" explicitly
    after adding an issue with the error details to the database.

    In case the wrapped function call succeeds, the decorator will pass on the
    resulting tuple unchanged.

    In case a known `UpdateTaskError` (or subclasses of it) exception is caught,
    the decorator will return the tuple of (`okay`, `fail`) int values, just as
    the decorated function is expected to do. Otherwise it will return (`-1`,
    `-1`) to indicate the count of updated / failed items is unknown.

    Parameters
    ----------
    func : function
        The "update" task function to be executed. It is expected to return a
        tuple of (`okay`, `fail`) int values.

    Returns
    -------
    (int, int)
        A tuple of (`okay`, `fail`) values as described above.
    """

    @wraps(func)
    def wrapper(*args, **kwargs):
        task_name = f"{func.__module__}.{func.__name__}"
        log.debug(f"Starting task: {task_name}")
        try:
            d_task = Task.objects.get(name=task_name)
            d_task.t_start = make_aware(datetime.now())
        except Task.DoesNotExist:
            d_task = Task(name=task_name, t_start=make_aware(datetime.now()))

        # use the first line of the function's docstring as task description:
        d_task.description = func.__doc__.split("\n")[0]
        d_task.save()

        try:
            result = func(*args, **kwargs)

        except (UpdateMachinesError, UpdateSessionsError) as err:
            d_task.t_end = make_aware(datetime.now())
            d_task.failed += 1
            d_task.save()
            d_issue = Issue(
                severity=Issue.Level.HIGH,
                task=d_task,
                description=str(err),
            )
            log.warning(f"❗📒  Registered issue {d_issue} in DB.")
            return err.okay, err.fail

        except Exception as err:
            d_task.failed += 1
            d_task.save()
            d_issue = Issue(
                severity=Issue.Level.HIGH,
                task=d_task,
                description=str(err),
            )
            d_issue.save()
            log.warning(f"❗📒  Registered issue {d_issue} in DB.")
            return -1, -1

        log.debug(f"Completed task: {task_name}")
        d_task.t_end = make_aware(datetime.now())
        d_task.failed = 0
        d_task.save()

        return result

    return wrapper


def fire_and_dont_forget(func=None, *, severity=Issue.Level.MEDIUM):
    """Run a function silencing exceptions (adding issue records if applicable).

    The decorated "one-shot" function will be called with any exception raised
    during the function's runtime being silenced and turned into a new record
    in the Issues table.

    Parameters
    ----------
    func : function
        The "one-shot" function that should be tracked.
    severity : Issue.Level, optional
        The level to be assigned to the issue if an exception has been raised
        by the decorated function, by default `Issue.Level.MEDIUM`.
    """
    if func is None:
        return partial(fire_and_dont_forget, severity=severity)

    @wraps(func)
    def wrapper(*args, **kwargs):
        task_name = f"{func.__module__}.{func.__name__}"
        log.debug(f"Tracking one-shot function: {task_name}")

        try:
            result = func(*args, **kwargs)

        except Exception as err:
            d_issue = Issue(
                severity=severity,
                description=f"{func.__name__}: {str(err)}\n\n{traceback.format_exc()}",
            )
            d_issue.save()
            log.warning(f"❗📒  Registered issue {d_issue} in DB.")
            return None

        log.debug(f"Completed one-shot function: {task_name}")

        return result

    return wrapper



"""Diop util methods."""

from smtplib import SMTPException
from sys import stderr

from django.conf import settings
from django.core.mail import EmailMessage
from django.utils.timezone import make_naive

from loguru import logger as log

from . import site
from .models.message_template import MessageTemplate


def naive_minutes(dt_obj):
    """Format a `datetime` to minute precision after running `make_naive` on it.

    Parameters
    ----------
    dt_obj : datetime.datetime
        The datetime object to format.
    """
    return make_naive(dt_obj).strftime(r"%Y-%m-%d %H:%M")


def send_email(
    subject: str, body: str, to: str, support: str = "", prefix: bool = True
):
    """Send an email.

    Simplifies things like deriving the sender address, putting a site-label
    prefix into the subject and dealing with cc/bcc-ing administrators.

    Parameters
    ----------
    subject : str
        The email subject.
    body : str
        The email body / content.
    to : str
        A single recipient address.
    support : str, optional
        If set to `cc` or `bcc` the address set in `DIOP_SUPPORT_CONTACT`
        will be added as a respective recipient. By default empty.
    prefix : bool, optional
        If set to `True` a site-specific label will be prefixed to the subject,
        enclosed in square brackets, otherwise the subject will not be touched
        (by default True).

    Raises
    ------
    SMTPException
        Raised in case sending the email failed.
    """

    cc = bcc = []
    if support == "cc":
        cc = [settings.DIOP_SUPPORT_CONTACT]
    elif support == "bcc":
        bcc = [settings.DIOP_SUPPORT_CONTACT]

    email = EmailMessage(
        subject=subject,
        body=body,
        from_email=f"{site.name()} Support <{settings.DIOP_SUPPORT_CONTACT}>",
        to=[to],
        cc=cc,
        bcc=bcc,
    )
    try:
        email.send(fail_silently=False)
    except SMTPException as err:
        log.error(f"Sending email to [{to}] failed: {err.with_traceback}")
        raise err


def send_support_email(template: str, context: dict):
    """Shorthand to send a mail to the `DIOP_SUPPORT_CONTACT` address.

    Parameters
    ----------
    template : str
        The name of the template to use for the message.
    context : dict
        The context dict required to render the template.
    """
    support = settings.DIOP_SUPPORT_CONTACT
    msg = MessageTemplate.render_message(template, context)
    send_email(msg.subject, msg.body, to=support)


def set_loglevel(level: str = "TRACE"):
    """Shorthand to adjust the logging level going to stderr.

    Mostly (but not exclusively) useful when interactively testing functionality
    using the `shell_plus` extension.

    Parameters
    ----------
    level : str, optional
        The level name, by default "TRACE".
    """
    log.remove()
    log.add(stderr, level=level)



"""Adapter module to pull data from an ActiveDirectory."""

import ldap

from django.conf import settings

from box import Box
from loguru import logger as log

from ..models import User


class ActiveDirectory:
    """A connection wrapper to fetch information from an ActiveDirectory (AD).

    Attributes
    ----------
    ldap_uri : str
        The address of the LDAP / AD server to connect to.
    bind_dn : str
        The bind DN ("bind user name").
    bind_pw : str
        The bind credentials ("bind user password").
    base_dn : str
        The base DN to use as the `scope` parameter for ActiveDirectory search
        queries, for example in `get_group_members()`.
    user_details : dict
        A dict with details on user accounts queried from ActiveDirectory, with
        usernames as keys.
    ad_group : str
        The name of the AD group being used by all group-related operations.
    conn : LDAPObject
        The LDAP / AD connection object.
    """

    def __init__(
        self, bind_dn=None, bind_pw=None, ldap_uri=None, ad_group=None, base_dn=None
    ):
        """ActiveDirectory constructor.

        Parameters
        ----------
        bind_dn : str, optional
            The bind DN ("bind user name"). If not specified
            `diop_core.settings.AUTH_LDAP_BIND_DN` will be used.
        bind_pw : str, optional
            The bind credentials ("bind user password"), if not specified
            `diop_core.settings.AUTH_LDAP_BIND_PASSWORD` will be used.
        ldap_uri : str, optional
            The address of the LDAP / AD server. Will default to
            `diop_core.settings.AUTH_LDAP_SERVER_URI` if not specified.
        ad_group : str, optional
            The name of the group to get the members for. Defaults to
            `diop_core.settings.DIOP_USER_AD_GROUP` if not specified.
        base_dn : str, optional
            The base DN to use in search queries. If not given explicitly,
            `diop_core.settings.AUTH_LDAP_BASE_DN` will be used.

        Raises
        ------
        Exception
            Any exception raised during the "bind" action will be re-raised.
        """
        self.bind_dn = settings.AUTH_LDAP_BIND_DN if not bind_dn else bind_dn
        self.bind_pw = settings.AUTH_LDAP_BIND_PASSWORD if not bind_pw else bind_pw
        self.ldap_uri = settings.AUTH_LDAP_SERVER_URI if not ldap_uri else ldap_uri
        self.ad_group = settings.DIOP_USER_AD_GROUP if not ad_group else ad_group
        self.base_dn = settings.AUTH_LDAP_BASE_DN if not base_dn else base_dn

        self._ad_group_filter = "(&(objectClass=GROUP)(cn={group_name}))"
        self._ad_user_filter_person = "(&(objectClass=USER)(objectCategory=person))"

        self._group_members = []
        self._group_members_usernames = []
        self.user_details = {}

        log.debug("Setting GLOBAL LDAP options from 'AUTH_LDAP_GLOBAL_OPTIONS'...")
        for opt_name, opt_setting in settings.AUTH_LDAP_GLOBAL_OPTIONS.items():
            ldap.set_option(opt_name, opt_setting)

        log.debug(f"Trying to connect to [{self.ldap_uri}]...")
        conn = ldap.initialize(self.ldap_uri)
        conn.protocol_version = 3
        conn.set_option(ldap.OPT_REFERRALS, 0)
        try:
            conn.simple_bind_s(who=self.bind_dn, cred=self.bind_pw)
        except Exception as err:
            log.error(f"LDAP bind failed: {err}")
            raise err

        self.conn = conn
        log.debug(f"Successfully connected to [{self.ldap_uri}].")

    @property
    def group_members(self) -> list:
        """A list of user-DN's being member of the configured AD group."""
        if not self._group_members:
            self._group_members = self.get_group_members(self.ad_group)

        return self._group_members

    @property
    def group_members_usernames(self) -> list:
        """A list of usernames being member of the configured AD group."""
        if not self._group_members_usernames:
            self._group_members_usernames = self.get_group_members_usernames(
                self.group_members
            )

        return self._group_members_usernames

    def get_group_members(self, group):
        """Query the AD to retrieve a list of user-DN's being member of a group.

        If any of the DN's in the list looks like a group-DN instead of a
        user-DN (contains the sequence "Group") a recursive lookup is attempted
        by using the value of the DN's first element as the group name. In case
        the recursive lookup fails, that DN is omitted and a warning message is
        issued to the log.

        Parameters
        ----------
        group : str
            The name of the group to get the members for, e.g. `XY-VDI-Users`.

        Returns
        -------
        list(str)
            A list of user DN's.
        """
        filter = self._ad_group_filter.replace("{group_name}", group)
        result = self.conn.search_s(self.base_dn, ldap.SCOPE_SUBTREE, filter)
        raw_members = result[0][1]["member"]

        members = []
        for member in raw_members:
            try:
                # sanitize by force-decoding (assuming it is UTF-8):
                member = member.decode("utf-8")
            except:
                pass

            if "Group" not in member:
                members.append(member)
                continue

            log.debug(f"Potential group: {member}")
            try:
                group_name = member.split(",")[0].split("=")[1]
                members += self.get_group_members(group_name)
            except Exception as err:
                log.warning(f"Unable to resolve [{member}]: {err}")

        log.debug(f"Got {len(members)} members for group [{group}].")
        return members

    def get_group_members_usernames(self, group_members_dn) -> list:
        """Fetch usernames for the given list of group member DN's.

        Parameters
        ----------
        group_members_dn : list(str)
            The group members DN's to fetch the usernames for.

        Returns
        -------
        list
        """
        usernames = []
        for user_dn in group_members_dn:
            details = self.user_details_from_dn(user_dn)
            if details:
                usernames.append(details.username)
        log.debug(f"Got {len(usernames)} usernames.")
        return usernames

    def user_details_from_dn(self, user_dn) -> Box:
        """Fetch display name, email and department for a user DN.

        Parameters
        ----------
        user_dn : str or str-like
            The user-DN to fetch details for.

        Returns
        -------
        Box
            A Box with the attributes listed below, or `None` in case the lookup
            failed:

            - `username` : str
            - `display_name` : str
            - `email` : str
            - `department` : str
            - `enabled` : bool
        """
        try:
            # sanitize by force-decoding (assuming it is UTF-8):
            user_dn = user_dn.decode("utf-8")
        except:
            pass

        log.debug(f"Getting details for [{user_dn}]...")
        result = self.conn.search_s(
            user_dn, ldap.SCOPE_BASE, self._ad_user_filter_person
        )
        if len(result) < 1:
            log.warning(f"No results for user DN [{user_dn}]!")
            return None

        if len(result) > 1:
            log.error(f"Huh?! Got multiple results for user DN [{user_dn}]!")
            return None

        details = result[0][1]
        display_name = details["displayName"][0].decode("utf-8")
        email = details["mail"][0].decode("utf-8")
        username = details["sAMAccountName"][0].decode("utf-8")
        user_account_control = details["userAccountControl"][0].decode("utf-8")
        enabled = True if not int(user_account_control) & 0x002 else False
        try:
            department = details["department"][0].decode("utf-8")
        except:  # ruff: noqa: E722 (bare-except)
            department = ""

        details = Box(
            {
                "username": username,
                "display_name": display_name,
                "email": email,
                "department": department,
                "enabled": enabled,
            }
        )
        # store in the object's dict `user_details` using `username` as the key:
        self.user_details[username] = details

        return details



"""Diop site related functions."""

from django.conf import settings
from django.contrib.sites.models import Site

from loguru import logger as log


def name():
    """Get the configured 'Site Name'."""
    return Site.objects.get(pk=settings.SITE_ID).name


def booking_uri():
    """Get the URI of the PPMS web interface for the configured core facility."""
    # FIXME: should rather be in the adapter module (to allow for potentially
    # other booking system adapters) but this currently leads to a circular
    # import - for now the function stays here, reconsider at a later point.
    pumapi = str(settings.PYPPMS_URI)
    cf_ref = settings.PYPPMS_CORE_REF
    return pumapi.replace("pumapi/", f"login/?pf={cf_ref}")


def in_dry_run_mode(msg: str) -> bool:
    """Check if "dry-run" mode is active and log a warning message if yes.

    This is intended as a shorthand to limit the amount of code necessary for
    evaluating if the `DIOP_DRY_RUN` setting is true and issue a warning to the
    log with a situation-specific message.

    Parameters
    ----------
    msg : str
        The caller-specific description to be added to the warning log message.

    Returns
    -------
    bool
    """
    if not settings.DIOP_DRY_RUN:
        return False

    log.warning(f"[🫧DRY-RUN🫧]: {msg} will be skipped...")
    return True



"""Wrapper replacing Django's default `manage.py` file.

Examples
--------
Run the Diop application using the built-in server (development only).
```bash
SETTINGS_ENV=diop-dev.env diop-manage runserver
```

Run the [`shell_plus` Django extension][shellplus]:
```bash
SETTINGS_ENV=diop-dev.env diop-manage shell_plus
```

[shellplus]: https://django-extensions.readthedocs.io/en/latest/shell_plus.html
"""

import os
import sys

from random import choices
from string import ascii_letters, digits
from textwrap import dedent


def manage():
    """Entry-point for running administrative tasks (replaces `manage.py`)."""
    os.environ.setdefault("DJANGO_SETTINGS_MODULE", "diop_core.settings")
    from django.core.management import execute_from_command_line

    execute_from_command_line(sys.argv)


def newconfig():
    """Print a config file template, pre-filled with some reasonable values."""
    secret = "".join(choices(ascii_letters + digits, k=64))
    config = f"""
    # Django secret key, used for session data, cookies and temporary tokens.
    # Changing the key will invalidate those, but won't affect user passwords.
    SECRET_KEY={secret}

    # SECURITY WARNING: don't run with debug turned on in production!
    DEBUG=False

    ALLOWED_HOSTS=127.0.0.1,localhost,diop.vdi.example.xy,diop-core.lxd

    DATABASE_URL=mysql://diop:diop_db_password@diop-db.lxd:3306/diop

    AUTH_LDAP_SERVER_URI=ldaps://dc.domainname.ads.example.xy
    AUTH_LDAP_BIND_DN=ldapbinduser@domainname.ads.example.xy
    AUTH_LDAP_BIND_PASSWORD=LdApBiNdPaSsWoRd
    AUTH_LDAP_BASE_DN=DC=domainname,DC=ads,DC=example,DC=xy

    EMAIL_BACKEND=django.core.mail.backends.console.EmailBackend
    EMAIL_HOST=not.relevant.for.console.backend

    PYPPMS_API_KEY=WhatAWonderfulKey
    PYPPMS_URI=https://ppms.eu/example/pumapi/
    PYPPMS_CACHE=/tmp/pyppms_cache/
    PYPPMS_CORE_REF=2
    PYPPMS_LOC_FILTER=vdi.example.xy
    PYPPMS_NAME_GROUPS=Argon,Helium,Neon


    # ResTricks server address:
    PSYTRICKS_BASE_URL="http://localhost:8080/"
    # Verify ResTricks server version when connecting:
    PSYTRICKS_VERIFY=True


    DIOP_USER_AD_GROUP=EXMPL-VDI-Users
    DIOP_SUPPORT_CONTACT=vdi-support@example.xy
    DIOP_DISCONNECTED_MAX=75
    DIOP_NTDOMAIN=EXAMPLE

    # Switching on "graceful" mode will make several tasks run in a less impacting
    # way, e.g. removing access to a DG will be done such that the standard group
    # will still be allowed to connect. Most likely not useful in production!
    DIOP_GRACEFUL=True

    # Switching on "dry-run" mode will prevent certain actions from doing effective
    # changes to the platform, i.e. changing access permissions or terminating
    # sessions. It will NOT affect components fetching status information to update
    # the DB, or strictly informational effects like e-mails or pop-up messages
    # (they will run normally). Actions that will be skipped due to dry-run being
    # active will print a warning-level message to the log instead.
    DIOP_DRY_RUN=False"""
    print(dedent(config))



"""Create Diop default groups and permissions."""

from django.contrib.auth.models import Group, Permission
from django.core.management.base import BaseCommand, CommandError


class Command(BaseCommand):
    """Command to create default groups and set up their permissions."""

    help = "Create Diop default groups and set up their permissions"

    @staticmethod
    def setup_group_admins():
        """Set up group `Admins`.

        The group will simply have **all** permissions assigned.
        """
        print("Adding auth group 'Admins'...")
        group, created = Group.objects.get_or_create(name="Admins")
        for newperm in Permission.objects.all():
            print(f"Adding permission [{newperm}]...")
            group.permissions.add(newperm)

    @staticmethod
    def setup_group(name, natural_keys):
        """Set up a group and assign permissions as defined.

        Parameters
        ----------
        name : str
            The name of the group. Note: the group may already exist, in which
            case only the permissions will be set accordingly.
        natural_keys : list
            A list of 3-tuples (`app_label`, `model`, `codename`) to be used as
            parameters for the `Permission.objects.get_by_natural_key()` call.
            The permissions defined in this list will be assigned to the group,
            any permission not present in the list will be removed!
        """
        new_permissions = []
        print(f"Adding auth group '{name}'...")
        group, created = Group.objects.get_or_create(name=name)
        for natural_key in natural_keys:
            newperm = Permission.objects.get_by_natural_key(
                app_label=natural_key[1],
                model=natural_key[2],
                codename=natural_key[0],
            )
            print(f"Adding permission [{newperm}]...")
            group.permissions.add(newperm)
            new_permissions.append(newperm)

        for permission in group.permissions.all():
            if permission not in new_permissions:
                print(f"Removing unexpected 🦹 permission [{permission}]...")
                group.permissions.remove(permission)

    def handle(self, *args, **options):
        """Perform group creation and permission assignment.

        Note
        ----
        To figure out how a specific 3-tuple is called, the following code may
        be used in a `shell_plus` session:

        ```Python
        for perm in Permission.objects.all():
            if "template" in perm.natural_key()[0]:
                print(perm.natural_key())
        ```
        """
        Command.setup_group_admins()

        operators_permissions = [
            # "view" permissions:
            ("view_logentry", "admin", "logentry"),
            ("view_group", "auth", "group"),
            ("view_permission", "auth", "permission"),
            ("view_user", "auth", "user"),
            ("view_contenttype", "contenttypes", "contenttype"),
            ("view_deliverygroup", "diop", "deliverygroup"),
            ("view_issue", "diop", "issue"),
            ("view_machine", "diop", "machine"),
            ("view_messagetemplate", "diop", "messagetemplate"),
            ("view_notification", "diop", "notification"),
            ("view_ppmsbooking", "diop", "ppmsbooking"),
            ("view_reservation", "diop", "reservation"),
            ("view_session", "diop", "session"),
            ("view_sessionproperties", "diop", "sessionproperties"),
            ("view_sitelog", "diop", "sitelog"),
            ("view_task", "diop", "task"),
            ("view_user", "diop", "user"),
            ("view_failure", "django_q", "failure"),
            ("view_ormq", "django_q", "ormq"),
            ("view_schedule", "django_q", "schedule"),
            ("view_success", "django_q", "success"),
            ("view_task", "django_q", "task"),
            ("view_session", "sessions", "session"),
            ("view_site", "sites", "site"),
            # "add" permissions:
            ### ("add_task", "diop", "task"),  # do NOT assign!
            ("add_task", "django_q", "task"),
            ("add_messagetemplate", "diop", "messagetemplate"),
            # "change" permissions:
            ("change_messagetemplate", "diop", "messagetemplate"),
            # "delete" permissions:
            ("delete_messagetemplate", "diop", "messagetemplate"),
            ("delete_failure", "django_q", "failure"),  # delete Failed task
            ("delete_ormq", "django_q", "ormq"),  # delete Queued task
        ]
        Command.setup_group(name="Operators", natural_keys=operators_permissions)

        users_permissions = [
            ("view_contenttype", "contenttypes", "contenttype"),
            ("view_session", "sessions", "session"),
            ("view_deliverygroup", "diop", "deliverygroup"),
            ("view_machine", "diop", "machine"),
            ("view_session", "diop", "session"),
            ("view_sessionproperties", "diop", "sessionproperties"),
        ]
        Command.setup_group(name="Users", natural_keys=users_permissions)



# Generated by Django 4.1.7 on 2023-04-14 10:43

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("diop", "0001_initial"),
    ]

    operations = [
        migrations.CreateModel(
            name="Machine",
            fields=[
                (
                    "fqdn",
                    models.CharField(max_length=64, primary_key=True, serialize=False),
                ),
                ("state", models.CharField(max_length=16)),
                ("powerstate", models.CharField(max_length=16)),
                ("maintenance", models.BooleanField(default=False)),
                ("registration", models.CharField(max_length=16)),
            ],
        ),
    ]



"""Custom management command for registering schdeuled tasks."""

from datetime import datetime, timedelta

from django.db import IntegrityError
from django.core.management.base import BaseCommand, CommandError
from django.utils import timezone

from django_q.models import Schedule


def next_time_at(hour, minute=0):
    """Get the next timezone-aware timepoint with given hour / minute.

    The function is creating a timezone-aware `datetime` object with the given
    `hour` and `minute` (all smaller components will be set to zero), making it
    the closest one that is still in the future.

    Meaning if that time has not yet passed today, the returned object will
    represent a timepoint of today, otherwise of tomorrow.

    Parameters
    ----------
    hour : int
        The hour for the timepoint.
    minute : int, optional
        The minute for the timepoint, by default 0.

    Returns
    -------
    datetime
    """
    tp = timezone.make_aware(
        datetime.now().replace(hour=hour, minute=minute, second=0, microsecond=0)
    )
    if tp < timezone.now():
        tp = tp + timedelta(days=1)

    return tp


class Command(BaseCommand):
    """Management command to register default Diop tasks in the scheduler."""

    help = "Register standard tasks in the scheduler"

    @staticmethod
    def remove_task(name):
        """Remove a task with the given name from the scheduler.

        Parameters
        ----------
        name : str
            The name of the task to be removed.
        """
        deleted, _ = Schedule.objects.filter(name=name).delete()
        if deleted:
            print(f"Removed task: '{name}'")

    @staticmethod
    def add_task_unless_existing(
        name,
        func,
        hook=None,
        args=None,
        kwargs=None,
        schedule_type=Schedule.ONCE,
        minutes=None,
        repeats=-1,
        next_run=timezone.now(),
        cron=None,
        cluster=None,
        intended_date_kwarg=None,
    ):
        """Add a task unless the Schedule already has one with the same name."""
        if Schedule.objects.filter(name=name).exists():
            print(f"Task already existing: '{name}'")
            return

        print(f"Registering new task: '{name}' ...")
        schedule = Schedule(
            name=name,
            func=func,
            hook=hook,
            args=args,
            kwargs=kwargs,
            schedule_type=schedule_type,
            minutes=minutes,
            repeats=repeats,
            next_run=next_run,
            cron=cron,
            cluster=cluster,
            intended_date_kwarg=intended_date_kwarg,
        )
        # trigger validation of the schedule
        schedule.full_clean()
        schedule.save()
        print(f"✅ Done registering new task: '{name}'")

    def handle(self, *args, **options):
        """Perform all task registrations in the scheduler.

        Any task is registered by first removing it from the scheduler and then
        re-adding it with the parameters defined here. This ensures that
        changing parameters here will update the corresponding tasks in the
        scheduler as well.

        Note
        ----
        Subclasses of `BaseCommand` must override their `handle` method. The
        method is called automatically after the Django environment has been
        configured and the CLI arguments have been parsed.
        """
        Command.remove_task(name="PPMS: Pull Users (Fast)")
        Command.add_task_unless_existing(
            func="diop.tasks.pull_ppms_users",
            name="PPMS - Pull Users Fast (@1m)",
            schedule_type=Schedule.MINUTES,
            minutes=1,
            cluster="ppms-short",
        )

        Command.remove_task(name="PPMS: Pull Users (Force-Refresh)")
        Command.add_task_unless_existing(
            func="diop.tasks.pull_ppms_users",
            name="PPMS - Pull Users Force-Refresh (@1h)",
            schedule_type=Schedule.HOURLY,
            cluster="ppms-long",
            kwargs={"force_refresh": True},
        )

        Command.remove_task(name="PPMS: Pull Bookings (2d)")
        Command.add_task_unless_existing(
            func="diop.tasks.pull_ppms_bookings",
            name="PPMS - Pull Bookings next 2d (@1m)",
            schedule_type=Schedule.MINUTES,
            minutes=1,
            cluster="ppms-short",
            kwargs={"days": 2},
        )

        Command.remove_task(name="PPMS: Pull Bookings (10d)")
        Command.add_task_unless_existing(
            func="diop.tasks.pull_ppms_bookings",
            name="PPMS - Pull Bookings next 10d (@1h)",
            schedule_type=Schedule.HOURLY,
            cluster="ppms-short",
            kwargs={"days": 10},
        )

        Command.remove_task(name="ActiveDirectory: Pull User Details")
        Command.add_task_unless_existing(
            func="diop.tasks.update_users_from_ad",
            name="ActiveDirectory - Pull User Details (@1h)",
            schedule_type=Schedule.HOURLY,
            cluster="activedirectory",
        )

        Command.remove_task(name="Housekeeping: @1m")
        Command.add_task_unless_existing(
            func="diop.tasks.housekeeping_every_minute",
            name="Housekeeping - Constantly (@1m)",
            schedule_type=Schedule.MINUTES,
            minutes=1,
            cluster="housekeeping",
        )

        Command.remove_task(name="Housekeeping: Day Start")
        Command.add_task_unless_existing(
            func="diop.tasks.housekeeping_start_of_day",
            name="Housekeeping - Day Start (09:00)",
            schedule_type=Schedule.DAILY,
            cluster="housekeeping",
            next_run=next_time_at(9),
        )

        Command.remove_task(name="Update Machines (@1m)")
        Command.add_task_unless_existing(
            func="diop.tasks.update_machine_status",
            name="Citrix - Pull Machine Status (@1m)",
            schedule_type=Schedule.MINUTES,
            minutes=1,
            cluster="statusupdates",
        )

        Command.remove_task(name="Update Sessions (@1m)")
        Command.add_task_unless_existing(
            func="diop.tasks.update_session_status",
            name="Citrix - Pull Session Status (@1m)",
            schedule_type=Schedule.MINUTES,
            minutes=1,
            cluster="statusupdates",
        )

        Command.remove_task(name="Disconnect Sessions")
        Command.add_task_unless_existing(
            func="diop.tasks.housekeeping_start_of_day",
            name="Housekeeping - Early Morning (04:00)",
            schedule_type=Schedule.DAILY,
            cluster="housekeeping",
            next_run=next_time_at(4),
        )
